@@ -1,8 +1,26 @@
 # -*- coding: utf-8 -*-
 
+# Copyright © 2009 Chris Jones <tortoise@tortuga>
+# Copyright © 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2011-2014 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 from __future__ import division, print_function, unicode_literals
 
-from traceback import format_tb
 import sys
 from Onboard.utils import unicode_str
 
