@@ -1,14 +1,39 @@
- # -*- coding: utf-8 -*-
+# -*- coding: utf-8 -*-
+
+# Copyright © 2008 Chris Jones <tortoise@tortuga>
+# Copyright © 2009-2011 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
+# Copyright © 2010-2015 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
 
 from __future__ import division, print_function, unicode_literals
 
 import time
+
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 from gi.repository import GObject, GLib, GdkX11, Gdk, Gtk
 
 from Onboard.definitions import DockingMonitor
-from Onboard.utils       import Rect, CallOnce, Timer, gtk_has_resize_grip_support
+from Onboard.utils       import Rect, Version
+from Onboard.Timer       import CallOnce, Timer
 from Onboard.WindowUtils import Orientation, WindowRectPersist, \
-                                set_unity_property
+                                set_unity_property, \
+                                gtk_has_resize_grip_support
 import Onboard.osk as osk
 
 ### Logging ###
@@ -58,7 +83,6 @@ class KbdWindowBase:
         self._current_docking_monitor = None
         self._monitor_workarea = {}
 
-        self._opacity = 1.0
         if gtk_has_resize_grip_support():
             self._default_resize_grip = self.get_has_resize_grip()
         else:
@@ -68,7 +92,6 @@ class KbdWindowBase:
         self._known_window_rects = []
         self._written_window_rects = {}
         self._written_dock_sizes = {}
-        self._wm_quirks = None
         self._auto_position_started = False
 
         self._desktop_switch_count = 0
@@ -133,25 +156,25 @@ class KbdWindowBase:
         >>> len(osk_util.get_current_wm_name().lower()) > 0
         True
         """
-        wm = config.quirks
+        wm = config.quirks_name
         if not wm:
             # Returns None on X error BadWindow (LP: 1016980)
             # Keep the same quirks as before in that case.
             wm = self._osk_util.get_current_wm_name()
 
         if wm:
-            self._wm_quirks = None
+            config.quirks = None
             for cls in [WMQuirksCompiz, WMQuirksMetacity, WMQuirksMutter]:
                 if  wm.lower() in cls.wms:
-                    self._wm_quirks = cls()
+                    config.quirks = cls()
                     break
 
-        if not self._wm_quirks:
-            self._wm_quirks = WMQuirksDefault()
+        if not config.quirks:
+            config.quirks = WMQuirksDefault()
 
         _logger.debug("window manager: {}".format(wm))
         _logger.debug("quirks selected: {}" \
-                                       .format(str(self._wm_quirks.__class__)))
+                                       .format(str(config.quirks.__class__)))
         return True
 
     def check_alpha_support(self):
@@ -174,11 +197,15 @@ class KbdWindowBase:
             self.set_visual(visual)
             self.keyboard_widget.set_visual(visual)
 
-            # full transparency for the window background
-            self.override_background_color(Gtk.StateFlags.NORMAL,
-                                           Gdk.RGBA(0, 0, 0, 0))
-            self.keyboard_widget.override_background_color(Gtk.StateFlags.NORMAL,
-                                           Gdk.RGBA(0, 0, 0, 0))
+            # Somehow Gtk 3.4 still needs these now deprecated calls,
+            # even though IconPalette and LabelPopups don't.
+            # Otherwise there will be a white background.
+            gtk_version = Version(Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION)
+            if gtk_version < Version(3, 18):  # Xenial doesn't need them
+                self.override_background_color(
+                    Gtk.StateFlags.NORMAL, Gdk.RGBA(0, 0, 0, 0))
+                self.keyboard_widget.override_background_color(
+                    Gtk.StateFlags.NORMAL, Gdk.RGBA(0, 0, 0, 0))
         else:
             _logger.info(_("no window transparency available;"
                            " screen doesn't support alpha channels"))
@@ -211,7 +238,7 @@ class KbdWindowBase:
         set_unity_property(self)
 
         if not config.xid_mode:   # not when embedding
-            ord = self.is_override_redirect_mode()
+            ord = config.is_override_redirect()
             if ord:
                 self.set_override_redirect(True)
 
@@ -232,7 +259,7 @@ class KbdWindowBase:
         if not config.xid_mode:   # not when embedding
             self.set_decorated(config.window.window_decoration)
 
-            type_hint = self._wm_quirks.get_window_type_hint(self)
+            type_hint = config.quirks.get_window_type_hint(self)
             self.set_type_hint(type_hint)
             self._type_hint = type_hint
 
@@ -247,12 +274,12 @@ class KbdWindowBase:
                 recreate = True
 
             # Override redirect toggled?
-            ord = self.is_override_redirect_mode()
+            ord = config.is_override_redirect()
             if ord != self.get_override_redirect():
                 recreate = True
 
             # Window type hint changed?
-            type_hint = self._wm_quirks.get_window_type_hint(self)
+            type_hint = config.quirks.get_window_type_hint(self)
             if type_hint != self._type_hint:
                 recreate = True
 
@@ -298,7 +325,7 @@ class KbdWindowBase:
                 self.icp.update_sticky_state()
 
     def update_taskbar_hint(self):
-        self._wm_quirks.update_taskbar_hint(self)
+        config.quirks.update_taskbar_hint(self)
 
     def is_visible(self):
         if not self.get_mapped():
@@ -318,7 +345,7 @@ class KbdWindowBase:
            self.can_move_into_view():
             self.move_home_rect_into_view()
 
-        self._wm_quirks.set_visible(self, visible)
+        config.quirks.set_visible(self, visible)
         self.on_visibility_changed(visible)
 
     def on_visibility_changed(self, visible):
@@ -351,23 +378,11 @@ class KbdWindowBase:
                 service.PropertiesChanged(service.IFACE,
                                           {'Visible': visible}, ['Visible'])
 
-    def set_opacity(self, opacity, force_set = False):
-        # Only set the opacity on visible windows.
-        # Metacity with compositing shows an unresponsive
-        # ghost of the window when trying to set opacity
-        # while it's hidden (LP: #929513).
-        _logger.debug("setting opacity to {}, force_set={}, "
-                      "visible={}" \
-                      .format(opacity, force_set, self.is_visible()))
-        if force_set:
-            Gtk.Window.set_opacity(self, opacity)
-        else:
-            if self.is_visible():
-                Gtk.Window.set_opacity(self, opacity)
-            self._opacity = opacity
+    def set_opacity(self, opacity):
+        self.keyboard_widget.set_opacity(opacity)
 
     def get_opacity(self):
-        return self._opacity
+        return self.keyboard_widget.get_opacity()
 
     def is_maximized(self):
         return self._maximized
@@ -389,11 +404,7 @@ class KbdWindowBase:
                 self.icp.hide()
 
     def _cb_visibility_notify(self, widget, event):
-        if not self.keyboard_widget.is_drag_initiated():
-            if event.state == Gdk.VisibilityState.UNOBSCURED:
-                # Metacity with compositing sometimes ignores set_opacity()
-                # immediately after unhiding. Set it here to be sure it sticks.
-                self.set_opacity(self._opacity)
+        pass
 
     def _cb_window_state_event(self, widget, event):
         """
@@ -536,7 +547,7 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         WindowRectPersist.__init__(self)
 
         GObject.signal_new("quit-onboard", KbdWindow,
-                           GObject.SIGNAL_RUN_LAST,
+                           GObject.SignalFlags.RUN_LAST,
                            GObject.TYPE_BOOLEAN, ())
 
         self._auto_position_poll_timer = Timer()
@@ -695,7 +706,7 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
             # In override redirect mode the window gets stuck with
             # the old scale -> recreate it
-            if self.is_override_redirect_mode() and \
+            if config.is_override_redirect() and \
                not startup and \
                not config.xid_mode:
                 self._recreate_window()
@@ -731,6 +742,19 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         if result == 0:
             self.update_home_rect()
 
+            # While moving with the title bar...
+            if config.has_window_decoration():
+
+                # stop auto-show from hiding the keyboard
+                keyboard = self.keyboard_widget.keyboard
+                keyboard.freeze_auto_show(1.0)
+
+                # cut off any leftover auto-show repositioning
+                self.stop_auto_positioning()
+
+                # keep the window active (LP: #1461823)
+                self.keyboard_widget.touch_inactivity_timer()
+
     def _filter_configure_event(self, rect):
         """
         Returns 0 for detected user positioning/sizing.
@@ -1009,7 +1033,7 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         #             correct configure-events
         if self._auto_position_started:
             self._auto_position_started = False
-            if self._wm_quirks.must_fix_configure_event():
+            if config.quirks.must_fix_configure_event():
                 self.move(x-1, y)
 
         self.move(x, y)
@@ -1395,9 +1419,6 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         rootwin = Gdk.get_default_root_window()
         return Rect.from_position_size(rootwin.get_position(),
                                 (rootwin.get_width(), rootwin.get_height()))
-    def is_override_redirect_mode(self):
-        return config.is_force_to_top() and \
-               self._wm_quirks.can_set_override_redirect(self)
 
     def assure_on_current_desktop(self):
         """
