@@ -1,4 +1,25 @@
 # -*- coding: UTF-8 -*-
+
+# Copyright © 2007 Martin Böhme <martin.bohm@kubuntu.org>
+# Copyright © 2008-2009 Chris Jones <tortoise@tortuga>
+# Copyright © 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2009, 2011-2014 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 """
 KeyCommon hosts the abstract classes for the various types of Keys.
 UI-specific keys should be defined in KeyGtk or KeyKDE files.
@@ -209,6 +230,9 @@ class KeyCommon(LayoutItem):
     # label that is currently displayed by this key
     label = ""
 
+    # mod_mask for the currently configured label
+    mod_mask = 0
+
     # smaller label of a currently invisible modifier level
     secondary_label = ""
 
@@ -292,9 +316,13 @@ class KeyCommon(LayoutItem):
         if label is None:
             label = ""
 
+        self.mod_mask = mod_mask
         self.label = label
         self.secondary_label = secondary_label
 
+        # Don't let erroneous labels shrink their whole size group.
+        self.ignore_group = label.startswith("0x")
+
     def draw_label(self, context = None):
         raise NotImplementedError()
 
@@ -314,6 +342,9 @@ class KeyCommon(LayoutItem):
     def get_id(self):
         return ""
 
+    def get_svg_id(self):
+        return ""
+
     def set_id(self, id, theme_id = None, svg_id = None):
         self.theme_id, self.id = self.parse_id(id)
         if theme_id:
@@ -413,8 +444,24 @@ class KeyCommon(LayoutItem):
 
     def is_return(self):
         id = self.id
-        return id == "RTRN" or \
-               id == "KPEN"
+        return (id == "RTRN" or
+                id == "KPEN")
+
+    def is_separator_cancelling(self):
+        """ Should this key cancel pending word separators? """
+        return (self.is_correction_key() or
+                self.is_return() or
+                self.id in set(["SPCE", "TAB",
+                                # Don't cancel for Backspace. We want to have
+                                # it appear to delete the pending separator.
+                                # This way it inserts a space, then immediately
+                                # deletes it.
+                                # "BKSP",
+                                "DELE",
+                                "LEFT", "RGHT", "UP", "DOWN",
+                                "HOME", "END", "PGUP", "PGDN",
+                                "INS", "ESC", "MENU",
+                                "Prnt", "Pause", "Scroll"]))
 
     def get_layer_index(self):
         assert(self.is_layer_button())
@@ -466,6 +513,19 @@ class RectKeyCommon(KeyCommon):
     def get_id(self):
         return self.id
 
+    def get_svg_id(self):
+        return self.svg_id
+
+    def get_state(self):
+        state = {}
+        state["prelight"]  = self.prelight
+        state["pressed"]   = self.pressed
+        state["active"]    = self.active
+        state["locked"]    = self.locked
+        state["scanned"]   = self.scanned
+        state["sensitive"] = self.sensitive
+        return state
+
     def draw(self, context = None):
         pass
 
@@ -860,7 +920,7 @@ class KeyPath:
         path._bounds = rect.copy()
         return path
 
-    _svg_path_pattern = re.compile("([+-]?[0-9.]*)")
+    _svg_path_pattern = re.compile("([+-]?[0-9.]+)")
 
     def copy(self):
         result = KeyPath()
