@@ -1,5 +1,26 @@
 # -*- coding: utf-8 -*-
 
+# Copyright © 2007 Martin Böhme <martin.bohm@kubuntu.org>
+# Copyright © 2007-2009 Chris Jones <tortoise@tortuga>
+# Copyright © 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
+# Copyright © 2009, 2011-2015 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 from __future__ import division, print_function, unicode_literals
 
 import sys
@@ -13,6 +34,8 @@ from gi.repository import Gdk
 import logging
 _logger = logging.getLogger(__name__)
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 try:
     from gi.repository import Atspi
 except ImportError as e:
@@ -26,11 +49,11 @@ from Onboard.Sound                 import Sound
 from Onboard.ClickSimulator        import ClickSimulator, \
                                           CSButtonMapper, CSFloatingSlave
 from Onboard.Scanner               import Scanner
-from Onboard.utils                 import Timer, ProgressiveDelayTimer, \
-                                          Modifiers, LABEL_MODIFIERS, \
+from Onboard.Timer                 import Timer, ProgressiveDelayTimer
+from Onboard.utils                 import Modifiers, LABEL_MODIFIERS, \
                                           parse_key_combination, \
                                           unicode_str
-from Onboard.definitions           import Handle
+from Onboard.definitions           import Handle, UIMask
 from Onboard.AutoShow              import AutoShow
 from Onboard.AutoHide              import AutoHide
 from Onboard.WordSuggestions       import WordSuggestions
@@ -47,16 +70,17 @@ config = Config()
 ########################
 
 
-# enum of event types for key press/release
 class EventType:
+    """ enum of event types for key press/release """
     (
         CLICK,
         DOUBLE_CLICK,
         DWELL,
     ) = range(3)
 
-# enum dock mode
+
 class DockMode:
+    """ enum dock mode """
     (
         FLOATING,
         BOTTOM,
@@ -64,19 +88,6 @@ class DockMode:
     ) = range(3)
 
 
-class UIMask:
-    (
-        CONTROLLERS,
-        SUGGESTIONS,
-        LAYOUT,
-        LAYERS,
-        SIZE,
-        REDRAW,
-    ) = (1<<bit for bit in range(6))
-
-    ALL = -1
-
-
 class UnpressTimers:
     """
     Redraw keys unpressed after a short while.
@@ -197,13 +208,13 @@ class KeySynthVirtkey(KeySynth):
         if self._vk:
             self._vk.release_keycode(keycode)
 
-    def lock_mod(self, mod):
+    def lock_mod(self, mod_mask):
         if self._vk:
-            self._vk.lock_mod(mod)
+            self._vk.lock_mod(mod_mask)
 
-    def unlock_mod(self, mod):
+    def unlock_mod(self, mod_mask):
         if self._vk:
-            self._vk.unlock_mod(mod)
+            self._vk.unlock_mod(mod_mask)
 
     def press_key_string(self, keystr):
         """
@@ -273,8 +284,6 @@ class TextChanger():
         self._key_synth_virtkey = KeySynthVirtkey(vk)
         self._key_synth_atspi = KeySynthAtspi(vk)
 
-        self._delayed_mods = None
-
         if config.keyboard.key_synth: # == KeySynth.ATSPI:
             self._key_synth = self._key_synth_atspi
         else: # if config.keyboard.key_synth == KeySynth.VIRTKEY:
@@ -319,60 +328,28 @@ class TextChanger():
     def release_keysym(self, keysym):
         self._key_synth.release_keysym(keysym)
 
-    def delay_modifier_changes(self):
-        """
-        Delay changing modifiers until apply_modifier_changes().
-        """
-        self._delayed_mods = {}
-
-    def apply_modifier_changes(self):
-        """
-        Apply accumulated modifiers changes.
-        """
-        if not self._delayed_mods is None:
-            for mod, count in self._delayed_mods.items():
-                if count:
-                    self._do_lock_mod(mod)
-                else:
-                    self._do_unlock_mod(mod)
-
-            self._delayed_mods = None
-
     def lock_mod(self, mod):
-        if self._delayed_mods is None:
-            self._do_lock_mod(mod)
-        else:
-            self._delayed_mods[mod] = True
-
-    def unlock_mod(self, mod):
-        if self._delayed_mods is None:
-            self._do_unlock_mod(mod)
-        else:
-            self._delayed_mods[mod] = False
-
-    def _do_lock_mod(self, mod):
         self._key_synth.lock_mod(mod)
 
-    def _do_unlock_mod(self, mod):
+    def unlock_mod(self, mod):
         self._key_synth.unlock_mod(mod)
 
     # Higher-level functions
-    def press_keysyms(self, key_name, count = 1):
+    def press_keysyms(self, key_name, count=1):
         """
         Generate any number of full key-strokes for the given named key symbol.
         """
         keysym = get_keysym_from_name(key_name)
         for i in range(count):
-            self.press_keysym  (keysym)
+            self.press_keysym(keysym)
             self.release_keysym(keysym)
 
-    def insert_string_at_caret(self, text):
+    def insert_string_at_caret(self, text, allow_insertion=True):
         """
-        Send key presses for all characters in a unicode string
-        and keep track of the changes in text_context.
+        Insert text at the caret position.
         """
         text_context = self.get_keyboard().text_context
-        if text_context.can_insert_text():
+        if allow_insertion and text_context.can_insert_text():
             text = text.replace("\\n", "\n")
             text_context.insert_text_at_caret(text)
         else:
@@ -489,7 +466,11 @@ class Keyboard(WordSuggestions):
         self._application = weakref.ref(application)
         self._pressed_key = None
         self._last_typing_time = 0
+
+        self._temporary_modifiers = None
+        self._locked_temporary_modifiers = None
         self._suppress_modifiers_stack = []
+        self._capitalization_requested = False
 
         self.layout = None
         self.scanner = None
@@ -596,6 +577,12 @@ class Keyboard(WordSuggestions):
         if layout_view in self._layout_views:
             self._layout_views.remove(layout_view)
 
+    def get_main_view(self):
+        layout_views = self._layout_views
+        if layout_views:
+            return layout_views[0]
+        return None
+
     def is_visible(self):
         for view in self._layout_views:
             visible = view.is_visible()
@@ -722,7 +709,8 @@ class Keyboard(WordSuggestions):
                    [BCMiddleClick, BCSingleClick, BCSecondaryClick,
                     BCDoubleClick, BCDragClick, BCHoverClick,
                     BCHide, BCShowClick, BCMove, BCPreferences, BCQuit,
-                    BCExpandCorrections, BCPauseLearning, BCLanguage,
+                    BCExpandCorrections, BCPreviousPredictions,
+                    BCNextPredictions, BCPauseLearning, BCLanguage,
                     BCStealthMode, BCAutoLearn, BCAutoPunctuation, BCInputline,
                    ]
                 }
@@ -804,12 +792,13 @@ class Keyboard(WordSuggestions):
         return key and \
                key.is_text_changing()
 
-    def key_down(self, key, view = None, sequence = None, action = True):
+    def key_down(self, key, view=None, sequence=None, action=True):
         """
         Press down on one of Onboard's key representations.
         This may be either an initial press, or a switch of the active_key
         due to dragging.
         """
+        self.on_any_key_down()
         self.stop_raise_attempts()
 
         if sequence:
@@ -817,7 +806,7 @@ class Keyboard(WordSuggestions):
             event_type = sequence.event_type
         else:
             button = 1
-            event_type =  EventType.CLICK
+            event_type = EventType.CLICK
 
         # Stop garbage collection delays until key release. They might cause
         # unexpected key repeats on slow systems.
@@ -830,6 +819,13 @@ class Keyboard(WordSuggestions):
             # stop timed redrawing for this key
             self._unpress_timers.stop(key)
 
+            # announce temporary modifiers
+            temp_mod_mask = 0
+            if config.keyboard.can_upper_case_on_button(button):
+                temp_mod_mask = Modifiers.SHIFT
+            self._set_temporary_modifiers(temp_mod_mask)
+            self._update_temporary_key_label(key, temp_mod_mask)
+
             # mark key pressed
             key.pressed = True
             self.on_key_pressed(key, view, sequence, action)
@@ -844,19 +840,17 @@ class Keyboard(WordSuggestions):
                 self._do_key_down_action(key, view, button, event_type)
 
             # remember as pressed key
-            if not key in self._pressed_keys:
+            if key not in self._pressed_keys:
                 self._pressed_keys.append(key)
 
-    def key_up(self, key, view = None, sequence = None, action = True):
+    def key_up(self, key, view=None, sequence=None, action=True):
         """ Release one of Onboard's key representations. """
-        update = False
-
         if sequence:
             button = sequence.button
             event_type = sequence.event_type
         else:
             button = 1
-            event_type =  EventType.CLICK
+            event_type = EventType.CLICK
 
         if key and \
            key.sensitive:
@@ -929,12 +923,13 @@ class Keyboard(WordSuggestions):
             self._non_modifier_released = False
             self._pressed_keys = []
             self._pressed_key = None
+            self.on_all_keys_up()
             gc.enable()
 
         # Process pending UI updates
         self.commit_ui_updates()
 
-    def key_long_press(self, key, view = None, button = 1):
+    def key_long_press(self, key, view=None, button=1):
         """ Long press of one of Onboard's key representations. """
         long_pressed = False
         key_type = key.type
@@ -974,22 +969,22 @@ class Keyboard(WordSuggestions):
                     long_pressed = True
 
         if long_pressed:
-            key.activated = True # no more drag selection
+            key.activated = True  # no more drag selection
 
         return long_pressed
 
     def _do_key_down_action(self, key, view, button, event_type):
+
         # generate key-stroke
         action = self.get_key_action(key)
-        can_send_key = (not key.sticky or not key.active) and \
-                        not key.action == KeyCommon.DELAYED_STROKE_ACTION and \
-                        not key.type == KeyCommon.WORD_TYPE
+        can_send_key = ((not key.sticky or not key.active) and
+                        not action == KeyCommon.DELAYED_STROKE_ACTION)
         if can_send_key:
             self.send_key_down(key, view, button, event_type)
 
         # Modifier keys may change multiple keys
         # -> redraw all dependent keys
-        # no danger of key repeats plus more work to do
+        # no danger of key repeats due to delays
         # -> redraw asynchronously
         if can_send_key and key.is_modifier():
             self.redraw_labels(False)
@@ -1005,7 +1000,7 @@ class Keyboard(WordSuggestions):
             if key.is_modifier() and \
                not self._can_cycle_modifiers():
                 can_send_key = True
-            else: # single touch/click
+            else:  # single touch/click
                 can_send_key = self.step_sticky_key(key, button, event_type)
 
             if can_send_key:
@@ -1013,7 +1008,7 @@ class Keyboard(WordSuggestions):
                 if key.is_modifier():
                     self.redraw_labels(False)
         else:
-            self.release_non_sticky_key(key, view, button, event_type)
+            self._release_non_sticky_key(key, view, button, event_type)
 
         # Multi-touch: temporarily stop cycling modifiers if
         # a non-modifier key was pressed. This way we get both,
@@ -1025,34 +1020,30 @@ class Keyboard(WordSuggestions):
     def send_key_down(self, key, view, button, event_type):
         if self.is_key_disabled(key):
             _logger.debug("send_key_down: "
-                          "rejecting blacklisted key action for '{}'" \
+                          "rejecting blacklisted key action for '{}'"
                           .format(key.id))
             return
 
-        key_type = key.type
         modifier = key.modifier
 
-        if modifier == 8: # Alt
+        if modifier == Modifiers.ALT and \
+           self._is_alt_special():
             self._last_alt_key = key
         else:
             action = self.get_key_action(key)
             if action != KeyCommon.DELAYED_STROKE_ACTION:
                 WordSuggestions.on_before_key_press(self, key)
-                self.maybe_send_alt_press_for_key(key, view, button, event_type)
+                self._maybe_send_alt_press_for_key(key, view,
+                                                   button, event_type)
+                self._maybe_lock_temporary_modifiers(key)
 
                 self.send_key_press(key, view, button, event_type)
 
-            if action == KeyCommon.DOUBLE_STROKE_ACTION: # e.g. CAPS
+            if action == KeyCommon.DOUBLE_STROKE_ACTION:  # e.g. CAPS
                 self.send_key_release(key, view, button, event_type)
 
         if modifier:
-            # Increment this before lock_mod() to skip
-            # updating keys a second time in set_modifiers().
-            self.mods[modifier] += 1
-
-            # Alt is special because it activates the window manager's move mode.
-            if modifier != 8: # not Alt?
-                self._text_changer.lock_mod(modifier)
+            self._do_lock_modifier(modifier)
 
             # Update word suggestions on shift press.
             self.invalidate_context_ui()
@@ -1069,19 +1060,33 @@ class Keyboard(WordSuggestions):
 
         key_type = key.type
         modifier = key.modifier
+        action = self.get_key_action(key)
 
-        if modifier == 8: # Alt
+        # Unlock most modifiers before key release, otherwise Compiz wall
+        # plugin's viewport switcher window doesn't close after
+        # Alt+Ctrl+Up/Down (LP: #1532254).
+        if modifier and \
+           action != KeyCommon.DOUBLE_STROKE_ACTION: # not NumLock, CAPS
+            self._do_unlock_modifier(modifier)
+
+            # Update word suggestions on shift unlatch or release.
+            self.invalidate_context_ui()
+
+        # generate key event(s)
+        if modifier == Modifiers.ALT and \
+           self._is_alt_special():
             pass
         else:
-            action = self.get_key_action(key)
             if action == KeyCommon.DOUBLE_STROKE_ACTION or \
                action == KeyCommon.DELAYED_STROKE_ACTION:
 
                 WordSuggestions.on_before_key_press(self, key)
-                self.maybe_send_alt_press_for_key(key, view, button, event_type)
+                self._maybe_send_alt_press_for_key(key, view,
+                                                   button, event_type)
+                self._maybe_lock_temporary_modifiers(key)
 
                 if key_type == KeyCommon.CHAR_TYPE:
-                    # allow to use AT-SPI direct text insertion for char keys
+                    # allow  direct text insertion by AT-SPI for char keys
                     self._text_changer.insert_string_at_caret(key.code)
                 else:
                     self.send_key_press(key, view, button, event_type)
@@ -1089,19 +1094,17 @@ class Keyboard(WordSuggestions):
             else:
                 self.send_key_release(key, view, button, event_type)
 
-        if modifier:
-            # Decrement before unlock_mod() to skip updating
-            # keys a second time in set_modifiers().
-            self.mods[modifier] -= 1
-
-            # Alt is special because it activates window manager's move mode.
-            if modifier != 8: # not Alt?
-                self._text_changer.unlock_mod(modifier)
+        # Unlock NumLock, CAPS, etc. after key events were sent,
+        # else they are toggled right back on.
+        if modifier and \
+           action == KeyCommon.DOUBLE_STROKE_ACTION:
+            self._do_unlock_modifier(modifier)
 
             # Update word suggestions on shift unlatch or release.
             self.invalidate_context_ui()
 
-        self.maybe_send_alt_release_for_key(key, view, button, event_type)
+        self._maybe_unlock_temporary_modifiers()
+        self._maybe_send_alt_release_for_key(key, view, button, event_type)
 
         # Check modifier counts for plausibility.
         # There might be a bug lurking that gets certain modifers stuck
@@ -1120,15 +1123,75 @@ class Keyboard(WordSuggestions):
                 # Reset this too, else unlatching won't happen until restart.
                 self._external_mod_changes[mod] = 0
 
-    def maybe_send_alt_press_for_key(self, key, view, button, event_type):
+    def _update_temporary_key_label(self, key, temp_mod_mask):
+        """ update label for temporary modifiers """
+        mod_mask = self.get_mod_mask()
+        temp_mod_mask |= mod_mask
+        if key.mod_mask != temp_mod_mask:
+            key.configure_label(temp_mod_mask)
+
+    def _set_temporary_modifiers(self, mod_mask):
+        """ Announce the intention to lock these modifiers on key-press. """
+        # only some single modifiers supported at this time
+        if not mod_mask or \
+           mod_mask in (Modifiers.SHIFT, Modifiers.CAPS, Modifiers.CTRL,
+                        Modifiers.SUPER, Modifiers.ALTGR):
+            self._temporary_modifiers = mod_mask
+
+    def _maybe_lock_temporary_modifiers(self, key):
+        """ Lock modifier before a single key-press """
+        modifier = self._temporary_modifiers
+        if modifier and \
+           not self.mods[modifier] and \
+           not key.modifier == modifier and \
+           not key.is_button():
+            self._locked_temporary_modifiers = modifier
+            self._do_lock_modifier(modifier)
+
+    def _maybe_unlock_temporary_modifiers(self):
+        """ Unlock modifier after a single key-press """
+        if self._locked_temporary_modifiers:
+            self._do_unlock_modifier(self._locked_temporary_modifiers)
+            self._locked_temporary_modifiers = None
+
+    def _do_lock_modifier(self, modifier):
+        """ Lock modifier in response to modifier presses. """
+        # Increment this before lock_mod() to skip
+        # updating keys a second time in set_modifiers().
+        self.mods[modifier] += 1
+
+        # Alt is special because it activates the window manager's move mode.
+        if modifier != Modifiers.ALT or \
+           not self._is_alt_special():  # not Alt?
+            self._text_changer.lock_mod(modifier)
+
+    def _do_unlock_modifier(self, modifier):
+        """ Unlock modifier in response to modifier releases. """
+        # Decrement before unlock_mod() to skip updating
+        # keys a second time in set_modifiers().
+        self.mods[modifier] -= 1
+
+        # Alt is special because it activates window manager's move mode.
+        if modifier != Modifiers.ALT or \
+            not self._is_alt_special(): # not Alt?
+            self._text_changer.unlock_mod(modifier)
+
+    def _is_alt_special(self):
+        """
+        Does the ALT key need special treatment due to it
+        """
+        return not config.is_override_redirect()
+
+    def _maybe_send_alt_press_for_key(self, key, view, button, event_type):
         """ handle delayed Alt press """
         if self.mods[8] and \
+           self._is_alt_special() and \
            not key.active and \
-           not self.is_key_disabled(key) and \
-           not key.type == KeyCommon.BUTTON_TYPE:
+           not self.is_key_disabled(key):
+           not key.type == KeyCommon.BUTTON_TYPE and \
             self.maybe_send_alt_press(view, button, event_type)
 
-    def maybe_send_alt_release_for_key(self, key, view, button, event_type):
+    def _maybe_send_alt_release_for_key(self, key, view, button, event_type):
         """ handle delayed Alt release """
         if self._alt_locked:
             self.maybe_send_alt_release(view, button, event_type)
@@ -1183,6 +1246,12 @@ class Keyboard(WordSuggestions):
         elif key_type == KeyCommon.SCRIPT_TYPE:
             activated = False
 
+        elif key_type == KeyCommon.WORD_TYPE:
+            activated = False
+
+        elif key_type == KeyCommon.CORRECTION_TYPE:
+            activated = False
+
         key.activated = activated
 
     def send_key_release(self, key, view, button = 1, event_type = EventType.CLICK):
@@ -1225,8 +1294,8 @@ class Keyboard(WordSuggestions):
 
     def _edit_snippet(self, view, snippet_id):
         if not config.xid_mode and \
-            not self.editing_snippet and \
-            view:
+           not self.editing_snippet and \
+           view:
             view.show_snippets_dialog(snippet_id)
             self.editing_snippet = True
 
@@ -1237,13 +1306,11 @@ class Keyboard(WordSuggestions):
             return True
         return False
 
-    def release_non_sticky_key(self, key, view, button, event_type):
-        # Delay locking/unlockign modifiers. Multiple locks and unlocks
-        # of SHIFT may happen when the punctuator is active. Make sure
-        # we set the final resulting modifier state only once, else we can't
-        # distinguish our changes from physical keyboard actions in
-        # set_modifiers.
-        self._text_changer.delay_modifier_changes()
+    def _release_non_sticky_key(self, key, view, button, event_type):
+        # Request capitalization before keys are unlatched, so we can
+        # prevent modifiers from toggling more than once and confuse
+        # set_modifiers().
+        WordSuggestions.on_before_key_release(self, key)
 
         # release key
         self.send_key_up(key, view, button, event_type)
@@ -1252,11 +1319,20 @@ class Keyboard(WordSuggestions):
         # keep them unchanged until the actual click happens.
         # -> allow clicks with modifiers
         if not key.is_layer_button() and \
-           not (key.type == KeyCommon.BUTTON_TYPE and \
+           not (key.type == KeyCommon.BUTTON_TYPE and
                 key.is_click_type_key()) and \
-           not key in self.get_text_displays():
+           key not in self.get_text_displays():
+
+            # Don't release SHIFT if we're going to enable
+            # capitalization anyway.
+            except_keys = None
+            if self._capitalization_requested:
+                except_keys = [key for key in self._latched_sticky_keys
+                               if key.modifier == Modifiers.SHIFT]
+
             # release latched modifiers
-            self.release_latched_sticky_keys(only_unpressed = True)
+            self.release_latched_sticky_keys(only_unpressed=True,
+                                             except_keys=except_keys)
 
             # undo temporary suppression of the text display
             WordSuggestions.show_input_line_on_key_release(self, key)
@@ -1273,8 +1349,53 @@ class Keyboard(WordSuggestions):
         # punctuation assistance and collapse corrections
         WordSuggestions.on_after_key_release(self, key)
 
-        # lock/unlock modifiers
-        self._text_changer.apply_modifier_changes()
+        # capitalization requested by punctuator?
+        if self._capitalization_requested:
+            self._capitalization_requested = False
+            if not self.mods[Modifiers.SHIFT]:  # SHIFT not active yet?
+                self._enter_caps_mode()
+
+    def request_capitalization(self, capitalize):
+        """
+        Request entering upper-caps mode after next key-release.
+        """
+        self._capitalization_requested = capitalize
+
+    def _enter_caps_mode(self):
+        """
+        Do what has to be done so that the next pressed
+        character will be capitalized.
+
+        Don't call key_down+up for this, because modifiers may be
+        configured not to latch.
+        """
+        lfsh_keys = self.find_items_from_ids(["LFSH"])
+        rtsh_keys = self.find_items_from_ids(["RTSH"])
+
+        # unlatch all shift keys
+        for key in rtsh_keys + lfsh_keys:
+            if key.active:
+                key.active = False
+                key.locked = False
+                if key in self._latched_sticky_keys:
+                    self._latched_sticky_keys.remove(key)
+                if key in self._locked_sticky_keys:
+                    self._locked_sticky_keys.remove(key)
+            self.redraw([key])
+
+        # Latch right shift for capitalization,
+        # if there is no right shift latch left shift instead.
+        shift_keys = rtsh_keys if rtsh_keys else lfsh_keys
+        for key in shift_keys:
+            if not key.active:
+                key.active = True
+                if key not in self._latched_sticky_keys:
+                    self._latched_sticky_keys.append(key)
+                self.redraw([key])
+
+        self.mods[Modifiers.SHIFT] = 1
+        self._text_changer.lock_mod(1)
+        self.redraw_labels(False)
 
     def maybe_switch_to_first_layer(self, key):
         """
@@ -1293,7 +1414,7 @@ class Keyboard(WordSuggestions):
                 self.active_layer_index = 0
                 self.invalidate_visible_layers()
                 self.invalidate_canvas()
-                self.invalidate_context_ui() # update layer button state
+                self.invalidate_context_ui()  # update layer button state
 
             return unlatch
 
@@ -1310,7 +1431,7 @@ class Keyboard(WordSuggestions):
     def set_modifiers(self, mod_mask):
         """
         Sync Onboard with modifiers of the given modifier mask.
-        Used to sync changes to system modifier state with Onboard.
+        Used to sync changes of system modifier state with Onboard.
         """
         # The special handling of ALT in Onboard confuses the detection of
         # modifier presses from the outside.
@@ -1321,9 +1442,17 @@ class Keyboard(WordSuggestions):
         if self._alt_locked:
             return
 
-        for mod_bit in (1<<bit for bit in range(8)):
+        if self._temporary_modifiers:
+            return
+
+        # SHIFT doesn't unlatch in Firefox, launchpad question entry, typing
+        # "?" after inserting "collection" with Small layout, Xenial,
+        if self.is_typing():
+            return
+
+        for mod_bit in (1 << bit for bit in range(8)):
             # Directly redraw locking modifiers only. All other modifiers
-            # redraw after a short delay. This is meant to prevents
+            # redraw after a short delay. This is meant to prevent
             # Onboard from busily flashing keys and using CPU while
             # typing with a hardware keyboard.
             if (mod_bit & (Modifiers.CAPS | Modifiers.NUMLK)):
@@ -1375,11 +1504,11 @@ class Keyboard(WordSuggestions):
             # re-draw delayed?
             if active and \
                draw_delay > 0.0:
-                self._queue_pending_redraw(mod_bit, active, keys, draw_delay)
+                self._queue_pending_modifier_redraw(mod_bit, active, keys, draw_delay)
             else:
                 self._redraw_modifier_keys(keys)
 
-    def _queue_pending_redraw(self, mod_bit, active, keys, delay):
+    def _queue_pending_modifier_redraw(self, mod_bit, active, keys, delay):
         item = self._pending_modifier_redraws.get(mod_bit)
         if item is None:
             # draw affected keys delayed
@@ -1401,7 +1530,7 @@ class Keyboard(WordSuggestions):
 
     def _redraw_pending_modifier_keys(self):
         for pending_active, keys in self._pending_modifier_redraws.values():
-             self._redraw_modifier_keys(keys)
+            self._redraw_modifier_keys(keys)
         self._pending_modifier_redraws = {}
 
     def _redraw_modifier_keys(self, keys):
@@ -1565,7 +1694,9 @@ class Keyboard(WordSuggestions):
                 if controller and \
                    controller.is_activated_on_press():
                     action = KeyCommon.SINGLE_STROKE_ACTION
-            else:
+
+            elif (key.type != KeyCommon.WORD_TYPE and
+                  key.type != KeyCommon.CORRECTION_TYPE):
                 label = key.get_label()
                 alternatives = self.find_canonical_equivalents(label)
                 if (len(label) == 1 and label.isalnum()) or \
@@ -1582,7 +1713,7 @@ class Keyboard(WordSuggestions):
 
         return action
 
-    def has_latched_sticky_keys(self, except_keys = None):
+    def has_latched_sticky_keys(self, except_keys=None):
         """ any sticky keys latched? """
         return len(self._latched_sticky_keys) > 0
 
@@ -1607,10 +1738,21 @@ class Keyboard(WordSuggestions):
                         if not skip_externally_set_modifiers or \
                            not key.is_modifier() or \
                            not self._external_mod_changes[key.modifier]:
-                            self.send_key_up(key)
-                            self._latched_sticky_keys.remove(key)
-                            key.active = False
-                            self.redraw([key])
+
+                            # Keep shift pressed if we're going to continue
+                            # upper-case anyway. Else multiple locks and
+                            # unlocks of SHIFT may happen when the punctuator
+                            # is active. We can change the modifier state at
+                            # most once per key release, else we can't
+                            # distinguish our changes from physical
+                            # keyboard actions in set_modifiers.
+                            if not key.modifier == Modifiers.SHIFT or \
+                               not self._capitalization_requested:
+
+                                self.send_key_up(key)
+                                self._latched_sticky_keys.remove(key)
+                                key.active = False
+                                self.redraw([key])
 
             # modifiers may change many key labels -> redraw everything
             self.redraw_labels(False)
@@ -1695,13 +1837,17 @@ class Keyboard(WordSuggestions):
 
         if mask & UIMask.CONTROLLERS:
             # update buttons
-            for controller in list(self.button_controllers.values()):
+            for controller in self.button_controllers.values():
                 controller.update()
-            mask = self._invalidated_ui  # may have changed by controllers
+            mask = self._invalidated_ui  # may have been changed by controllers
 
         if mask & UIMask.SUGGESTIONS:
             keys.update(WordSuggestions.update_suggestions_ui(self))
 
+            # update buttons that depend on suggestions
+            for controller in self.button_controllers.values():
+                controller.update_late()
+
         if mask & UIMask.LAYERS:
             self.update_visible_layers()
 
@@ -1752,29 +1898,34 @@ class Keyboard(WordSuggestions):
     def on_key_pressed(self, key, view, sequence, action):
         """ pressed state of a key instance was set """
         if sequence: # Not a simulated key press, scanner?
-            allowed = self.can_give_keypress_feedback()
+            feedback = self.can_give_keypress_feedback()
 
             # audio feedback
             if action and \
                config.keyboard.audio_feedback_enabled:
-                point = sequence.root_point \
-                        if allowed else (-1, -1) # keep passwords privat
-                Sound().play(Sound.key_feedback, *point)
+                pt = sequence.root_point \
+                     if feedback else (-1, -1) # keep passwords privat
+                pts = pt \
+                      if config.keyboard.audio_feedback_place_in_space \
+                      else (-1, -1)
+                Sound().play(Sound.key_feedback, pt[0], pt[1], pts[0], pts[1])
 
             # key label popup
             if not config.xid_mode and \
                config.keyboard.touch_feedback_enabled and \
                sequence.event_type != EventType.DWELL and \
                key.can_show_label_popup() and \
-               allowed:
+               feedback:
                 self._touch_feedback.show(key, view)
 
     def on_key_unpressed(self, key):
         """ pressed state of a key instance was cleard """
+        self._set_temporary_modifiers(0)
+        self._update_temporary_key_label(key, 0)
         self.redraw([key])
         self._touch_feedback.hide(key)
 
-    def on_outside_click(self):
+    def on_outside_click(self, button):
         """
         Called by outside click polling.
         Keep this as Francesco likes to have modifiers
@@ -1782,10 +1933,11 @@ class Keyboard(WordSuggestions):
         """
         self.release_latched_sticky_keys()
         self._click_sim.end_mapped_click()
+        WordSuggestions.on_outside_click(self, button)
 
     def on_cancel_outside_click(self):
         """ Called when outside click polling times out. """
-        pass
+        WordSuggestions.on_cancel_outside_click(self)
 
     def get_click_simulator(self):
         if config.mousetweaks and \
@@ -2001,12 +2153,16 @@ class ButtonController(object):
         """ asynchronous ui update """
         pass
 
+    def update_late(self):
+        """ after suggestions have been updated """
+        pass
+
     def can_dwell(self):
         """ can start dwelling? """
         return False
 
     def is_activated_on_press(self):
-        """ Cannot already called press() without consequences? """
+        """ Cannot cancel already called press() without consequences? """
         return False
 
     def set_visible(self, visible):
@@ -2225,7 +2381,7 @@ class BCMove(ButtonController):
                          Handle.MOVE in config.window.window_handles)
 
     def is_activated_on_press(self):
-        return True # cannot undo on press, dragging is already in progress
+        return True  # cannot undo on press, dragging is already in progress
 
 
 class BCLayer(ButtonController):
@@ -2297,7 +2453,7 @@ class BCQuit(ButtonController):
             GLib.idle_add(app.do_quit_onboard)
 
     def update(self):
-        self.set_visible(not config.xid_mode and \
+        self.set_visible(not config.xid_mode and
                          not config.lockdown.disable_quit)
 
 
@@ -2310,6 +2466,35 @@ class BCExpandCorrections(ButtonController):
         wordlist.expand_corrections(not wordlist.are_corrections_expanded())
 
 
+class BCPreviousPredictions(ButtonController):
+
+    id = "previous-predictions"
+
+    def release(self, view, button, event_type):
+        wordlist = self.key.get_parent()
+        wordlist.goto_previous_predictions()
+        self.keyboard.invalidate_context_ui()
+
+    def update_late(self):
+        wordlist = self.key.get_parent()
+        self.set_sensitive(wordlist.can_goto_previous_predictions())
+
+
+class BCNextPredictions(ButtonController):
+
+    id = "next-predictions"
+
+    def release(self, view, button, event_type):
+        wordlist = self.key.get_parent()
+        wordlist.goto_next_predictions()
+        self.keyboard.invalidate_context_ui()
+
+    def update_late(self):
+        key = self.key
+        wordlist = key.get_parent()
+        self.set_sensitive(wordlist.can_goto_next_predictions())
+
+
 class BCPauseLearning(ButtonController):
 
     id = "pause-learning"
