@@ -1,4 +1,24 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2009 Chris Jones <tortoise@tortuga>
+# Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
+# Copyright © 2009, 2011-2015 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 """ GTK keyboard widget """
 
 from __future__ import division, print_function, unicode_literals
@@ -7,6 +27,8 @@ import sys
 import time
 from math import sin, pi
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 from gi.repository          import GLib, Gdk, Gtk
 
 from Onboard.TouchInput     import TouchInput, InputSequence
@@ -18,7 +40,8 @@ from Onboard.KeyGtk         import Key
 from Onboard.KeyCommon      import LOD
 from Onboard.TouchHandles   import TouchHandles
 from Onboard.LayoutView     import LayoutView
-from Onboard.utils          import Rect, Timer, FadeTimer, escape_markup
+from Onboard.utils          import Rect, escape_markup
+from Onboard.Timer          import Timer, FadeTimer
 from Onboard.definitions    import Handle
 from Onboard.WindowUtils    import WindowManipulator, \
                                    canvas_to_root_window_rect, \
@@ -119,6 +142,7 @@ class InactivityTimer(Timer):
         self._keyboard.commit_transition()
         return False
 
+
 class HideInputLineTimer(Timer):
     """
     Temporarily hides the input line when the pointer touches it.
@@ -218,13 +242,18 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         LayoutView.__init__(self, keyboard)
         TouchInput.__init__(self)
 
+        self.set_app_paintable(True)
+
         self.canvas_rect = Rect()
+        self._opacity = 1.0
 
         self._last_click_time = 0
         self._last_click_key = None
 
         self._outside_click_timer = Timer()
         self._outside_click_detected = False
+        self._outside_click_num = 0
+        self._outside_click_button_mask = 0
         self._outside_click_start_time = None
 
         self._long_press_timer = Timer()
@@ -309,6 +338,16 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             self.touch_handles.set_window(win)
             self.update_window_handles()
 
+    def set_opacity(self, opacity):
+        """ Override deprecated Gtk function of the same name """
+        if self._opacity != opacity:
+            self._opacity = opacity
+            self.redraw()
+
+    def get_opacity(self):
+        """ Override deprecated Gtk function of the same name """
+        return self._opacity
+
     def set_startup_visibility(self):
         win = self.get_kbd_window()
         assert(win)
@@ -330,8 +369,8 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         # when inactive transparency is enabled.
         screen = self.get_screen()
         if screen and screen.is_composited() and \
-            self.inactivity_timer.is_enabled():
-            win.set_opacity(0.05, True) # keep it slightly visible just in case
+           self.inactivity_timer.is_enabled():
+            self.set_opacity(0.05)  # keep it slightly visible just in case
 
         # transition to initial opacity
         self.transition_visible_to(visible, 0.0, 0.4)
@@ -375,10 +414,10 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         rect = self.canvas_rect.deflate(self.get_frame_width())
 
-        layout.update_log_rect() # update logical tree to base aspect ratio
+        layout.update_log_rect()  # update logical tree to base aspect ratio
         rect = self._get_aspect_corrected_layout_rect(rect,
                                                       layout.context.log_rect)
-        layout.do_fit_inside_canvas(rect) # update contexts to final aspect
+        layout.do_fit_inside_canvas(rect)  # update contexts to final aspect
 
         # update the aspect ratio of the main window
         self.on_layout_updated()
@@ -621,8 +660,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         window = self.get_kbd_window()
         if window:
-            if window.get_opacity() != opacity:
-                window.set_opacity(opacity)
+            self.set_opacity(opacity)
 
             visible_before = window.is_visible()
             visible_later  = state.target_visibility
@@ -696,6 +734,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             self._outside_click_timer.start(0.01, self._on_click_timer)
             self._outside_click_detected = False
             self._outside_click_start_time = time.time()
+            self._outside_click_num = 0
 
     def stop_click_polling(self):
         self._outside_click_timer.stop()
@@ -706,15 +745,40 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         dunno, x, y, mask = rootwin.get_pointer()
         if mask & BUTTON123_MASK:
             self._outside_click_detected = True
+            self._outside_click_button_mask = mask
         elif self._outside_click_detected:
-            # button released anywhere outside of onboard's control
+            self._outside_click_detected = False
+            # A button was released anywhere outside of Onboard's control.
             _logger.debug("click polling: outside click")
-            self.stop_click_polling()
-            self.close_key_popup()
-            self.keyboard.on_outside_click()
-            return False
 
-        # stop after 30 seconds
+            self.close_key_popup()
+
+            button = \
+                self._get_button_from_mask(self._outside_click_button_mask)
+
+            # When clicking left, don't stop polling right away. This allows
+            # the user to select some text and paste it with middle click,
+            # while the pending separator is still inserted.
+            self._outside_click_num += 1
+            max_clicks = 4
+
+            if button != 1:  # middle and right click stop polling immediately
+                self.stop_click_polling()
+                self.keyboard.on_outside_click(button)
+
+            elif button == 1 and self._outside_click_num == 1:
+                if not config.wp.delayed_word_separators_enabled:
+                    self.stop_click_polling()
+                self.keyboard.on_outside_click(button)
+
+            # allow a couple of left clicks with delayed separators
+            elif self._outside_click_num >= max_clicks:
+                self.stop_click_polling()
+                self.keyboard.on_cancel_outside_click()
+
+            return True
+
+        # stop polling after 30 seconds
         if time.time() - self._outside_click_start_time > 30.0:
             self.stop_click_polling()
             self.keyboard.on_cancel_outside_click()
@@ -722,6 +786,15 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         return True
 
+    @staticmethod
+    def _get_button_from_mask(mask):
+        for i, bit in enumerate((Gdk.ModifierType.BUTTON1_MASK,
+                                 Gdk.ModifierType.BUTTON2_MASK,
+                                 Gdk.ModifierType.BUTTON3_MASK,)):
+            if mask & bit:
+                return i + 1
+        return 0
+
     def get_drag_window(self):
         """ Overload for WindowManipulator """
         return self.get_kbd_window()
@@ -790,6 +863,8 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             self.invalidate_for_resize(self._lod)
 
     def on_enter_notify(self, widget, event):
+        self.keyboard.on_activity_detected()
+
         self._update_double_click_time()
 
         # ignore event if a mouse button is held down
@@ -857,6 +932,8 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
     def on_input_sequence_begin(self, sequence):
         """ Button press/touch begin """
+        self.keyboard.on_activity_detected()
+
         self.stop_click_polling()
         self.stop_dwelling()
         self.close_key_popup()
@@ -1218,6 +1295,8 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
                 GLib.idle_add(self._on_touch_handles_opacity, 1.0, False)
 
     def _on_draw(self, widget, context):
+        context.push_group()
+
         decorated = LayoutView.draw(self, widget, context)
 
         # draw touch handles (enlarged move and resize handles)
@@ -1226,6 +1305,9 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             self.touch_handles.set_corner_radius(corner_radius)
             self.touch_handles.draw(context)
 
+        context.pop_group_to_source()
+        context.paint_with_alpha(self._opacity)
+
     def _overcome_initial_key_resistance(self, sequence):
         """
         Drag-select: Increase the hit area of the initial key
@@ -1304,7 +1386,15 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         rects = []
         for key in keys:
             r = key.get_canvas_border_rect()
-            rects.append(canvas_to_root_window_rect(self, r))
+            r = canvas_to_root_window_rect(self, r)
+
+            # scale coordinates in response to changes to
+            # org.gnome.desktop.interface scaling-factor
+            scale = config.window_scaling_factor
+            if scale and scale != 1.0:
+                r = r.scale(scale)
+
+            rects.append(r)
 
         return rects
 
@@ -1389,11 +1479,10 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             self._key_popup = None
 
     def _create_key_popup(self, parent):
-        popup = LayoutPopup(self.keyboard,
-                         self.close_key_popup)
+        popup = LayoutPopup(self.keyboard, self.close_key_popup)
         popup.supports_alpha = self.supports_alpha
         popup.set_transient_for(parent)
-        popup.set_opacity(parent.get_opacity())
+        popup.set_opacity(self.get_opacity())
         return popup
 
     def _show_key_popup(self, popup, key):
@@ -1543,7 +1632,13 @@ class KeyMenu:
         if self._closure:
             self._closure()
 
-    def _menu_positioning_func(self, menu, key):
+    def _menu_positioning_func(self, *params):
+        # work around change in number of paramters in Wily with Gtk 3.16
+        if len(params) == 4:
+            menu, x, y, key = params  # new in Wily
+        else:
+            menu, key = params
+
         r = self._keyboard_widget.get_key_screen_rect(key)
         menu_size  = (menu.get_allocated_width(),
                       menu.get_allocated_width())
@@ -1660,16 +1755,18 @@ class SuggestionMenu(KeyMenu):
         menu = Gtk.Menu()
 
         item = Gtk.MenuItem.new_with_mnemonic(_("_Remove suggestion…"))
-        item.connect("activate", self._on_remove_suggestion)
+        item.connect("activate", self._on_remove_suggestion, key)
         menu.append(item)
 
         return menu
 
-    def _on_remove_suggestion(self, menu_item):
+    def _on_remove_suggestion(self, menu_item, key):
         keyboard = self._keyboard
-        suggestion, history = keyboard.get_prediction_choice_and_history( \
-                                                           self._choice_index)
-        history = history[-1:] # only single word history supported
+        wordlist = key.get_parent()
+        suggestion, history = \
+            keyboard.get_prediction_choice_and_history(wordlist,
+                                                       self._choice_index)
+        history = history[-1:]  # only single word history supported
         dialog = RemoveSuggestionConfirmationDialog(
                     self._keyboard_widget.get_kbd_window(),
                     keyboard, suggestion, history)
@@ -1689,8 +1786,8 @@ class SuggestionMenu(KeyMenu):
         if menu_size[0] > rkey.w:
             return rkey.left(), rkey.bottom()
         else:
-            return super(SuggestionMenu, self).get_menu_position(
-                                                          rkey, menu_size)
+            return super(SuggestionMenu, self) \
+                .get_menu_position(rkey, menu_size)
 
 
 class RemoveSuggestionConfirmationDialog(Gtk.MessageDialog):
@@ -1710,8 +1807,8 @@ class RemoveSuggestionConfirmationDialog(Gtk.MessageDialog):
             self.set_transient_for(parent)
 
         # Don't hide dialog behind the keyboard in force-to-top mode.
-        #if config.is_force_to_top():
-        #    dialog.set_position(Gtk.WindowPosition.CENTER)
+        if config.is_force_to_top():
+            self.set_position(Gtk.WindowPosition.CENTER)
 
         markup = "<big>" + _("Remove word suggestion") + "</big>"
         markup = escape_markup(markup, preserve_tags=True)
