@@ -1,5 +1,24 @@
 # -*- coding: utf-8 -*-
 
+# Copyright © 2008-2009 Chris Jones <tortoise@tortuga>
+# Copyright © 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2011-2014 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 from __future__ import division, print_function, unicode_literals
 
 ### Logging ###
@@ -60,8 +79,11 @@ class LayoutLoaderSVG:
     # label_margin, theme_id, popup_id
     LAYOUT_FORMAT_3_1         = Version(3, 1)
 
+    # new key attributes show_active
+    LAYOUT_FORMAT_3_2         = Version(3, 2)
+
     # current format
-    LAYOUT_FORMAT             = LAYOUT_FORMAT_3_1
+    LAYOUT_FORMAT             = LAYOUT_FORMAT_3_2
 
     # precalc mask permutations
     _label_modifier_masks = permute_mask(LABEL_MODIFIERS)
@@ -88,8 +110,10 @@ class LayoutLoaderSVG:
         if layout:
             # purge attributes only used during loading
             for item in layout.iter_items():
-                item.templates = None
-                item.keysym_rules = None
+                if not item.templates is None:
+                    item.templates = None
+                if not item.keysym_rules is None:
+                    item.keysym_rules = None
 
             # enable caching
             layout = LayoutRoot(layout)
@@ -151,10 +175,12 @@ class LayoutLoaderSVG:
         for child in dom_node.childNodes:
             if child.nodeType == minidom.Node.ELEMENT_NODE:
 
-                # Skip over items with non-matching layout string.
+                # Skip over items with non-matching keyboard layout string.
                 # Items with the same id are processed from top to bottom,
                 # the first match wins. If no item matches we fall back to
                 # the item without layout string.
+                # This is used to select between alternative key definitions
+                # depending on the current system layout.
                 can_load = False
                 if not child.hasAttribute("id"):
                     can_load = True
@@ -164,12 +190,13 @@ class LayoutLoaderSVG:
                         if child.hasAttribute("layout"):
                             layout = child.attributes["layout"].value
                             can_load = self._has_matching_layout(layout)
+
+                            # don't look at items with this id again
+                            if can_load:
+                                loaded_ids.add(id)
                         else:
                             can_load = True
 
-                    if can_load:
-                        # don't look at items with this id again
-                        loaded_ids.add(id)
 
                 if can_load:
                     tag = child.tagName
@@ -307,10 +334,6 @@ class LayoutLoaderSVG:
         if value and value.lower() == 'false':
             item.scannable = False
 
-        value = attributes.get("unlatch_layer")
-        if not value is None:
-            item.unlatch_layer = value == "true"
-
         value = attributes.get("scan_priority")
         if not value is None:
             item.scan_priority = int(value)
@@ -354,7 +377,7 @@ class LayoutLoaderSVG:
         # find template attributes
         attributes = {}
         if node.hasAttribute("id"):
-            theme_id, id = RectKey.split_id(node.attributes["id"].value)
+            theme_id, id = RectKey.parse_id(node.attributes["id"].value)
             attributes.update(self.find_template(parent, RectKey, [id]))
 
         # let current node override any preceding templates
@@ -362,7 +385,7 @@ class LayoutLoaderSVG:
 
         # handle common layout-item attributes
         key = self._init_item(attributes, item_class)
-        key.parent = parent # assign early to have get_filename() work
+        key.parent = parent  # assign early to have get_filename() work
 
         # handle key-specific attributes
         self._init_key(key, attributes)
@@ -370,7 +393,7 @@ class LayoutLoaderSVG:
         # get key geometry from the closest svg file
         filename = key.get_filename()
         if not filename:
-            if not attribute.get("group") == "wsbutton":
+            if not attributes.get("group") == "wsbutton":
                 _logger.warning(_format("Ignoring key '{}'."
                                         " No svg filename defined.",
                                         key.id))
@@ -484,7 +507,7 @@ class LayoutLoaderSVG:
         if override:
             olabel, ogroup = override
             if olabel:
-                labels = { 0 : olabel[:]}
+                labels = {0 : olabel[:]}
                 if ogroup:
                     group_name = ogroup[:]
 
@@ -508,6 +531,15 @@ class LayoutLoaderSVG:
             if attributes["show_border"].lower() == 'false':
                 key.show_border = False
 
+        # show_active: allow to display key in latched or locked state
+        # legacy show_active behavior was hard-coded for layer0
+        if self._format < LayoutLoaderSVG.LAYOUT_FORMAT_3_2:
+            if key.id == "layer0":
+                key.show_active = False
+        if "show_active" in attributes:
+            if attributes["show_active"].lower() == 'false':
+                key.show_active = False
+
         if "label_x_align" in attributes:
             key.label_x_align = float(attributes["label_x_align"])
         if "label_y_align" in attributes:
@@ -566,17 +598,10 @@ class LayoutLoaderSVG:
         if key.type == KeyCommon.KEYCODE_TYPE and \
            not key.id in ["BKSP"]:
             if self._vk: # xkb keyboard found?
-                try:
-                    vkmodmasks = self._label_modifier_masks
-                    if sys.version_info.major == 2:
-                        vkmodmasks = [long(m) for m in vkmodmasks]
-                    vklabels = self._vk.labels_from_keycode(key.code,
-                                                            vkmodmasks)
-                except TypeError:
-                    # virtkey until 0.61.0 didn't have the extra param.
-                    vkmodmasks = (0, 1, 2, 128, 129) # used to be hard-coded
-                    vklabels = self._vk.labels_from_keycode(key.code)
-
+                vkmodmasks = self._label_modifier_masks
+                if sys.version_info.major == 2:
+                    vkmodmasks = [long(m) for m in vkmodmasks]
+                vklabels = self._vk.labels_from_keycode(key.code, vkmodmasks)
                 if sys.version_info.major == 2:
                     vklabels = [x.decode("UTF-8") for x in vklabels]
                 labels = {m : l for m, l in zip(vkmodmasks, vklabels)}
@@ -769,7 +794,7 @@ class LayoutLoaderSVG:
     def _has_matching_layout(self, layout_str):
         """
         Check if one ot the given layout strings matches
-        system layout and variant.
+        system keyboard layout and variant.
 
         Doctests:
         >>> l = LayoutLoaderSVG()
