@@ -1,17 +1,40 @@
 # -*- coding: UTF-8 -*-
 
+# Copyright © 2007 Martin Böhme <martin.bohm@kubuntu.org>
+# Copyright © 2007-2010 Chris Jones <tortoise@tortuga>
+# Copyright © 2008-2010 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2012-2013 Gerd Kohlberger <lowfi@chello.at>
+# Copyright © 2010-2015 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 from __future__ import division, print_function, unicode_literals
 
-### Logging ###
 import logging
 _logger = logging.getLogger("OnboardGtk")
-###############
 
 import sys
 import time
 import signal
 import os.path
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
+from gi.repository import GLib, Gdk, Gtk
+
 try:
     import dbus
     import dbus.service
@@ -19,9 +42,6 @@ try:
 except ImportError:
     pass
 
-
-from gi.repository import GLib, Gdk, Gtk
-
 from Onboard.KbdWindow       import KbdWindow, KbdPlugWindow
 from Onboard.Keyboard        import Keyboard
 from Onboard.KeyboardWidget  import KeyboardWidget
@@ -30,8 +50,9 @@ from Onboard.LayoutLoaderSVG import LayoutLoaderSVG
 from Onboard.Appearance      import ColorScheme
 from Onboard.IconPalette     import IconPalette
 from Onboard.Exceptions      import LayoutFileError
-from Onboard.utils           import show_confirmation_dialog, CallOnce, \
-                                    unicode_str
+from Onboard.utils           import unicode_str
+from Onboard.Timer           import CallOnce, Timer
+from Onboard.WindowUtils     import show_confirmation_dialog
 import Onboard.osk as osk
 
 ### Config Singleton ###
@@ -39,7 +60,7 @@ from Onboard.Config import Config
 config = Config()
 ########################
 
-import Onboard.KeyCommon
+from Onboard.KeyCommon import KeyCommon
 
 app = "onboard"
 DEFAULT_FONTSIZE = 10
@@ -63,23 +84,36 @@ class OnboardGtk(object):
         Gdk.set_program_class(app[0].upper() + app[1:])
 
         # no python3-dbus on Fedora17
-        if not "dbus" in globals():
-            _logger.warning("D-Bus bindings unavailable. "
-                            "Onboard will start with reduced functionality. "
-                            "Single-instance check, "
-                            "D-Bus service and "
-                            "hover click are disabled.")
+        bus = None
+        err_msg = ""
+        if "dbus" not in globals():
+            err_msg = "D-Bus bindings unavailable"
         else:
             # Use D-bus main loop by default
             dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
 
-            # Yield to GNOME Shell's keyboard before any other D-Bus activity
-            # to reduce the chance for D-Bus timeouts when enabling a11y keboard.
-            if not self._can_show_in_current_desktop():
-                sys.exit(0)
+            # Don't fail to start in Xenial's lightdm due
+            # to D-Bus not available.
+            try:
+                bus = dbus.SessionBus()
+            except dbus.exceptions.DBusException:
+                err_msg = "D-Bus session bus unavailable"
+                bus = None
 
-            # Check if there is already a Onboard instance running
-            bus = dbus.SessionBus()
+        if not bus:
+            _logger.warning(err_msg + "  " +
+                            "Onboard will start with reduced functionality. "
+                            "Single-instance check, "
+                            "D-Bus service and "
+                            "hover click are disabled.")
+
+        # Yield to GNOME Shell's keyboard before any other D-Bus activity
+        # to reduce the chance for D-Bus timeouts when enabling a11y keboard.
+        if not self._can_show_in_current_desktop(bus):
+            sys.exit(0)
+
+        if bus:
+            # Check if there is already an Onboard instance running
             has_remote_instance = bus.name_has_owner(self.DBUS_NAME)
 
             # Onboard in Ubuntu on first start silently embeds itself into
@@ -90,7 +124,7 @@ class OnboardGtk(object):
             # the user session (LP: 938302).
             if config.xid_mode and \
                config.launched_by == config.LAUNCHER_GSS and \
-               not (config.gnome_a11y and \
+               not (config.gnome_a11y and
                     config.gnome_a11y.screen_keyboard_enabled) and \
                not has_remote_instance:
                 sys.exit(0)
@@ -100,7 +134,8 @@ class OnboardGtk(object):
                 if has_remote_instance and \
                    not config.options.allow_multiple_instances:
                     # Present remote instance
-                    remote = bus.get_object(self.DBUS_NAME, ServiceOnboardKeyboard.PATH)
+                    remote = bus.get_object(self.DBUS_NAME,
+                                            ServiceOnboardKeyboard.PATH)
                     remote.Show(dbus_interface=ServiceOnboardKeyboard.IFACE)
                     _logger.info("Exiting: Not the primary instance.")
                     sys.exit(0)
@@ -134,6 +169,7 @@ class OnboardGtk(object):
         self._window = None
         self.status_icon = None
         self.service_keyboard = None
+        self._reload_layout_timer = Timer()
 
         # finish config initialization
         config.init()
@@ -164,17 +200,20 @@ class OnboardGtk(object):
             sys.stdout.write('%d\n' % self._window.get_id())
             sys.stdout.flush()
         else:
-            icp = IconPalette()
+            icp = IconPalette(self.keyboard)
             icp.set_layout_view(self.keyboard_widget)
             icp.connect("activated", self._on_icon_palette_acticated)
+            self.do_connect(icp.get_menu(), "quit-onboard",
+                        lambda x: self.do_quit_onboard())
 
             self._window = KbdWindow(self.keyboard_widget, icp)
             self.do_connect(self._window, "quit-onboard",
                             lambda x: self.do_quit_onboard())
 
-        #config.xid_mode = True
+        # config.xid_mode = True
         self._window.application = self
-        config.main_window = self._window # need this to access screen properties
+        # need this to access screen properties
+        config.main_window = self._window
 
         # load the initial layout
         _logger.info("Loading initial layout")
@@ -189,9 +228,9 @@ class OnboardGtk(object):
                 size = options.size.split("x")
                 rect.w = int(size[0])
                 rect.h = int(size[1])
-            if not options.x is None:
+            if options.x is not None:
                 rect.x = options.x
-            if not options.y is None:
+            if options.y is not None:
                 rect.y = options.y
 
             # Make sure the keyboard fits on screen
@@ -200,26 +239,26 @@ class OnboardGtk(object):
             if rect != self._window.get_rect():
                 orientation = self._window.get_screen_orientation()
                 self._window.write_window_rect(orientation, rect)
-                self._window.restore_window_rect() # move/resize early
+                self._window.restore_window_rect()  # move/resize early
 
         # export dbus service
         if not config.xid_mode and \
            "dbus" in globals():
-            self.service_keyboard = ServiceOnboardKeyboard(self.keyboard_widget)
+            self.service_keyboard = ServiceOnboardKeyboard(self)
 
         # show/hide the window
         self.keyboard_widget.set_startup_visibility()
 
         # keep keyboard window and icon palette on top of dash
-        if not config.xid_mode: # be defensive, not necessary when embedding
-            self._osk_util.keep_windows_on_top([self._window,
-                                                self._window.icp])
+        self._keep_windows_on_top()
 
         # connect notifications for keyboard map and group changes
         self.keymap = Gdk.Keymap.get_default()
-        self.do_connect(self.keymap, "keys-changed", self.cb_keys_changed) # map changes
+        # map changes
+        self.do_connect(self.keymap, "keys-changed", self.cb_keys_changed)
         self.do_connect(self.keymap, "state-changed", self.cb_state_changed)
-        Gdk.event_handler_set(cb_any_event, self)          # group changes
+        # group changes
+        Gdk.event_handler_set(cb_any_event, self)
 
         # connect config notifications here to keep config from holding
         # references to keyboard objects.
@@ -227,25 +266,30 @@ class OnboardGtk(object):
         reload_layout       = lambda x: once(self.reload_layout_and_present)
         update_ui           = lambda x: once(self._update_ui)
         update_ui_no_resize = lambda x: once(self._update_ui_no_resize)
-        update_transparency = lambda x: once(self.keyboard_widget.update_transparency)
+        update_transparency = \
+            lambda x: once(self.keyboard_widget.update_transparency)
         update_inactive_transparency = \
-                              lambda x: once(self.keyboard_widget.update_inactive_transparency)
+            lambda x: once(self.keyboard_widget.update_inactive_transparency)
 
         # general
-        config.auto_show.enabled_notify_add(lambda x: \
-                                    self.keyboard_widget.update_auto_show())
+        config.auto_show.enabled_notify_add(lambda x:
+                                    self.keyboard.update_auto_show())
+        config.auto_show.hide_on_key_press_notify_add(lambda x:
+                                    self.keyboard.update_auto_hide())
 
         # keyboard
         config.keyboard.key_synth_notify_add(reload_layout)
-        config.keyboard.input_event_source_notify_add(lambda x: \
+        config.keyboard.input_event_source_notify_add(lambda x:
                                     self.keyboard.update_input_event_source())
+        config.keyboard.touch_input_notify_add(lambda x:
+                                    self.keyboard.update_touch_input_mode())
         config.keyboard.show_secondary_labels_notify_add(update_ui)
 
-
         # window
-        config.window.window_state_sticky_notify_add(lambda x: \
-                                   self._window.update_sticky_state())
-        config.window.window_decoration_notify_add(self._on_window_options_changed)
+        config.window.window_state_sticky_notify_add(
+            lambda x: self._window.update_sticky_state())
+        config.window.window_decoration_notify_add(
+            self._on_window_options_changed)
         config.window.force_to_top_notify_add(self._on_window_options_changed)
         config.window.keep_aspect_ratio_notify_add(update_ui)
 
@@ -260,7 +304,7 @@ class OnboardGtk(object):
         config.layout_filename_notify_add(reload_layout)
 
         # theme
-        #config.gdi.gtk_theme_notify_add(self.on_gtk_theme_changed)
+        # config.gdi.gtk_theme_notify_add(self.on_gtk_theme_changed)
         config.theme_notify_add(self.on_theme_changed)
         config.key_label_font_notify_add(reload_layout)
         config.key_label_overrides_notify_add(reload_layout)
@@ -274,7 +318,7 @@ class OnboardGtk(object):
 
         # word suggestions
         config.word_suggestions.show_context_line_notify_add(update_ui)
-        config.word_suggestions.enabled_notify_add(lambda x: \
+        config.word_suggestions.enabled_notify_add(lambda x:
                                  self.keyboard.on_word_suggestions_enabled(x))
         config.word_suggestions.auto_learn_notify_add(
                                  update_ui_no_resize)
@@ -286,14 +330,14 @@ class OnboardGtk(object):
                                  self.keyboard.on_word_suggestions_enabled(x))
         config.word_suggestions.spelling_suggestions_enabled_notify_add(lambda x: \
                                  self.keyboard.on_spell_checker_changed())
+        config.word_suggestions.delayed_word_separators_enabled_notify_add(lambda x: \
+                                 self.keyboard.on_punctuator_changed())
         config.word_suggestions.wordlist_buttons_notify_add(
                                  update_ui_no_resize)
 
         # universal access
         config.scanner.enabled_notify_add(self.keyboard._on_scanner_enabled)
-
-        config.window.resize_handles_notify_add(lambda x: \
-                                 self.keyboard_widget.update_resize_handles())
+        config.window.window_handles_notify_add(self._on_window_handles_changed)
 
         # misc
         config.keyboard.show_click_buttons_notify_add(update_ui)
@@ -303,8 +347,8 @@ class OnboardGtk(object):
 
         # create status icon
         self.status_icon = Indicator()
-        self.status_icon.set_keyboard_window(self._window)
-        self.do_connect(self.status_icon, "quit-onboard",
+        self.status_icon.set_keyboard(self.keyboard)
+        self.do_connect(self.status_icon.get_menu(), "quit-onboard",
                         lambda x: self.do_quit_onboard())
 
         # Callbacks to use when icp or status icon is toggled
@@ -316,9 +360,17 @@ class OnboardGtk(object):
 
         # Minimize to IconPalette if running under GDM
         if 'RUNNING_UNDER_GDM' in os.environ:
+            _logger.info("RUNNING_UNDER_GDM set, turning on icon palette")
             config.icp.in_use = True
+            _logger.info("RUNNING_UNDER_GDM set, turning off indicator")
             config.show_status_icon = False
 
+            # For some reason the new values don't arrive in gsettings when
+            # running the unit test "test_running_in_live_cd_environment".
+            # -> Force gsettings to apply them, that seems to do the trick.
+            config.icp.apply()
+            config.apply()
+
         # unity-2d needs the skip-task-bar hint set before the first mapping.
         self.show_hide_taskbar()
 
@@ -412,7 +464,7 @@ class OnboardGtk(object):
 
     # Method concerning the icon palette
     def _on_icon_palette_acticated(self, widget):
-        self.keyboard_widget.toggle_visible()
+        self.keyboard.toggle_visible()
 
     def cb_icp_in_use_toggled(self, icp_in_use):
         """
@@ -459,18 +511,21 @@ class OnboardGtk(object):
 
         TODO would be nice if appeared to iconify to taskbar
         """
-        self.keyboard_widget.toggle_visible()
+        self.keyboard.toggle_visible()
 
+    def cb_group_changed(self):
+        """ keyboard group change """
+        self.reload_layout_delayed()
 
-    # keyboard layout changes
     def cb_keys_changed(self, keymap):
-        self.reload_layout()
+        """ keyboard map change """
+        self.reload_layout_delayed()
 
-    # modifier changes
     def cb_state_changed(self, keymap):
-        _logger.debug("keyboard state changed to 0x{:x}" \
-                      .format(keymap.get_modifier_state()))
+        """ keyboard modifier state change """
         mod_mask = keymap.get_modifier_state()
+        _logger.debug("keyboard state changed to 0x{:x}" \
+                      .format(mod_mask))
         self.keyboard.set_modifiers(mod_mask)
 
     def cb_vk_timer(self):
@@ -494,6 +549,10 @@ class OnboardGtk(object):
             self.keyboard.invalidate_ui_no_resize()
             self.keyboard.commit_ui_updates()
 
+    def _on_window_handles_changed(self, value = None):
+        self.keyboard_widget.update_window_handles()
+        self._update_ui()
+
     def _on_window_options_changed(self, value = None):
         self._update_window_options()
         self.keyboard.commit_ui_updates()
@@ -549,6 +608,21 @@ class OnboardGtk(object):
         config.apply_theme()
         self.reload_layout()
 
+    def _keep_windows_on_top(self, enable=True):
+        if not config.xid_mode: # be defensive, not necessary when embedding
+            if enable:
+                windows = [self._window, self._window.icp]
+            else:
+                windows = []
+            _logger.debug("keep_windows_on_top {}".format(windows))
+            self._osk_util.keep_windows_on_top(windows)
+
+    def on_focusable_gui_opening(self):
+        self._keep_windows_on_top(False)
+
+    def on_focusable_gui_closed(self):
+        self._keep_windows_on_top(True)
+
     def reload_layout_and_present(self):
         """
         Reload the layout and briefly show the window
@@ -557,6 +631,18 @@ class OnboardGtk(object):
         self.reload_layout(force_update = True)
         self.keyboard_widget.update_transparency()
 
+    def reload_layout_delayed(self):
+        """
+        Delay reloading the layout on keyboard map or group changes
+        This is mainly for LP #1313176 when Caps-lock is set up as
+        an accelerator to switch between keyboard "input sources" (layouts)
+        in unity-control-center->Text Entry (aka region panel).
+        Without waiting until after the shortcut turns off numlock,
+        the next "input source" (layout) is skipped and a second one
+        is selected.
+        """
+        self._reload_layout_timer.start(.5, self.reload_layout)
+
     def reload_layout(self, force_update=False):
         """
         Checks if the X keyboard layout has changed and
@@ -596,9 +682,9 @@ class OnboardGtk(object):
             self.vk_timer = GLib.timeout_add_seconds(1, self.cb_vk_timer)
 
     def load_layout(self, layout_filename, color_scheme_filename):
-        _logger.info("Loading keyboard layout from " + layout_filename)
+        _logger.info("Loading keyboard layout " + layout_filename)
         if (color_scheme_filename):
-            _logger.info("Loading color scheme from " + color_scheme_filename)
+            _logger.info("Loading color scheme " + color_scheme_filename)
 
         vk = self.get_vk()
 
@@ -606,12 +692,7 @@ class OnboardGtk(object):
                        if color_scheme_filename else None
         layout = LayoutLoaderSVG().load(vk, layout_filename, color_scheme)
 
-        self.keyboard.cleanup()
-        self.keyboard.set_virtkey(vk)
-        self.keyboard.layout = layout
-        self.keyboard.color_scheme = color_scheme
-        self.keyboard.on_layout_loaded()
-        self.keyboard_widget.on_layout_loaded()
+        self.keyboard.set_layout(layout, color_scheme, vk)
 
         if self._window and self._window.icp:
             self._window.icp.queue_draw()
@@ -645,6 +726,8 @@ class OnboardGtk(object):
         self.cleanup()
 
     def cleanup(self):
+        self._reload_layout_timer.stop()
+
         config.cleanup()
 
         # Make an effort to disconnect all handlers.
@@ -658,7 +741,7 @@ class OnboardGtk(object):
                 self.keyboard.scanner = None
             self.keyboard.cleanup()
 
-        self.status_icon.set_keyboard_window(None)
+        self.status_icon.set_keyboard(None)
         self._window.cleanup()
         self._window.destroy()
         self._window = None
@@ -668,7 +751,7 @@ class OnboardGtk(object):
         config.final_cleanup()
 
     @staticmethod
-    def _can_show_in_current_desktop():
+    def _can_show_in_current_desktop(bus):
         """
         When GNOME's "Typing Assistent" is enabled in GNOME Shell, Onboard
         starts simultaneously with the Shell's built-in screen keyboard.
@@ -684,16 +767,25 @@ class OnboardGtk(object):
         """
         result = True
 
+        # Content of XDG_CURRENT_DESKTOP:
+        # Before Vivid: GNOME Shell:   GNOME
+        #               GNOME Classic: GNOME
+        # Since Vivid:  GNOME Shell:   GNOME
+        #               GNOME Classic: GNOME-Classic:GNOME
+
         if config.options.not_show_in:
-            bus = dbus.SessionBus()
             current = os.environ.get("XDG_CURRENT_DESKTOP", "")
             names = config.options.not_show_in.split(",")
             for name in names:
-                if name == "GNOME":
-                    if bus.name_has_owner("org.gnome.Shell"):
-                        result = False
-                elif name == current:
-                    result = False
+                if name == current:
+                    # Before Vivid GNOME Shell and GNOME Classic had the same
+                    # desktop name "GNOME". Use the D-BUS name to decide if we
+                    # are in the Shell.
+                    if name == "GNOME":
+                        if bus and bus.name_has_owner("org.gnome.Shell"):
+                            result = False
+                    else:
+                        result  = False
 
             if not result:
                 _logger.info("Command line option not-show-in={} forbids running in "
@@ -715,9 +807,10 @@ if "dbus" in globals():
             _dbus_error_name = 'org.onboard.Exception'
 
 
-        def __init__(self, keyboard):
+        def __init__(self, app):
             super(ServiceOnboardKeyboard, self).__init__(dbus.SessionBus(), self.PATH)
-            self._keyboard = keyboard
+            self._keyboard = app.keyboard
+            self._keyboard_widget = app.keyboard_widget
 
         @dbus.service.method(dbus_interface=IFACE)
         def Show(self):
@@ -731,6 +824,22 @@ if "dbus" in globals():
         def ToggleVisible(self):
             self._keyboard.toggle_visible()
 
+        # private method, for unit-testing only
+        if config.is_running_from_source:
+            @dbus.service.method(dbus_interface=IFACE,
+                                 out_signature='a(sada{is}a{sb})')
+            def GetKeyState(self):
+                result = []
+                for key in self._keyboard.iter_keys():
+                    r = self._keyboard_widget.get_key_screen_rect(key)
+                    if not r.is_empty() and key.is_path_visible():
+                        labels = {m:l for m, l in key.labels.items() if l}  # -None
+                        result.append([key.get_id(),
+                                    list(r),
+                                    labels,
+                                    key.get_state()])
+                return result
+
         @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                              in_signature='ss', out_signature='v')
         def Get(self, iface, prop):
@@ -807,7 +916,7 @@ def cb_any_event(event, onboard):
     # XkbStateNotify maps to Gdk.EventType.NOTHING
     # https://bugzilla.gnome.org/show_bug.cgi?id=156948
     if type == Gdk.EventType.NOTHING:
-        onboard.reload_layout()
+        onboard.cb_group_changed()
 
     # Update the cached pango layout object here or Onboard
     # doesn't get those settings, i.e. label fonts sizes are off
