@@ -1,4 +1,9 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2013 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
 # Onboard is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 3 of the License, or
@@ -11,10 +16,6 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
-# Copyright © 2012, marmuta
-#
-# This file is part of Onboard.
 
 from __future__ import division, print_function, unicode_literals
 
@@ -34,10 +35,6 @@ class TextSpan:
     >>> span = TextSpan(3, 2, "0123456789")
     >>> span.get_span_text()
     '34'
-    >>> span.get_text_until_span()
-    '01234'
-    >>> span.get_text_from_span()
-    '3456789'
     """
 
     def __init__(self, pos = 0, length = 0, text = "", text_pos = 0):
@@ -129,6 +126,11 @@ class TextSpan:
         """
         Return the beginning of the whole available text,
         ending with and including the span.
+
+        Doctests:
+        >>> span = TextSpan(3, 2, "0123456789")
+        >>> span.get_text_until_span()
+        '01234'
         """
         return self.text[:self.end() - self.text_pos]
 
@@ -136,9 +138,25 @@ class TextSpan:
         """
         Return the end of the whole available text,
         starting from and including the span.
+
+        Doctests:
+        >>> span = TextSpan(3, 2, "0123456789")
+        >>> span.get_text_from_span()
+        '3456789'
         """
         return self.text[self.pos - self.text_pos:]
 
+    def get_text_after_span(self):
+        """
+        Return the remaining available text after the span.
+
+        Doctests:
+        >>> span = TextSpan(3, 2, "0123456789")
+        >>> span.get_text_after_span()
+        '56789'
+        """
+        return self.text[self.end() - self.text_pos:]
+
     def get_char_before_span(self):
         """
         Character right before the span.
@@ -159,6 +177,26 @@ class TextSpan:
         pos = self.pos - self.text_pos
         return self.text[pos - 1 : pos]
 
+    def get_last_char_in_span(self):
+        """
+        Character right before the span.
+
+        Doctests:
+        >>> span = TextSpan(0, 0, "0123456789", 0)
+        >>> span.get_last_char_in_span()
+        ''
+
+        >>> span = TextSpan(9, 1, "0123456789", 0)
+        >>> span.get_last_char_in_span()
+        '9'
+
+        >>> span = TextSpan(5, 2, "3456789", 3)
+        >>> span.get_last_char_in_span()
+        '6'
+        """
+        pos = self.end() - self.text_pos
+        return self.text[pos - 1 : pos]
+
     def _escape(self, text):
         return text.replace("\n", "\\n")
 
