@@ -1,4 +1,9 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2015 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
 # Onboard is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 3 of the License, or
@@ -11,19 +16,17 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
-# Copyright © 2012, marmuta
-#
-# This file is part of Onboard.
 
 from __future__ import division, print_function, unicode_literals
 
-import time
 import unicodedata
+import time
 
 import logging
 _logger = logging.getLogger(__name__)
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 try:
     from gi.repository import Atspi
 except ImportError as e:
@@ -32,7 +35,8 @@ except ImportError as e:
 from Onboard.AtspiStateTracker import AtspiStateTracker, AtspiStateType
 from Onboard.TextDomain        import TextDomains
 from Onboard.TextChanges       import TextChanges, TextSpan
-from Onboard.utils             import Timer, KeyCode, Modifiers
+from Onboard.utils             import KeyCode
+from Onboard.Timer             import Timer
 from Onboard                   import KeyCommon
 
 ### Config Singleton ###
@@ -60,10 +64,10 @@ class TextContext:
     def insert_text_at_caret(self, text):
         return NotImplementedError()
 
-    def delete_text(self, offset, length = 1):
+    def delete_text(self, offset, length=1):
         return NotImplementedError()
 
-    def delete_text_before_caret(self, length = 1):
+    def delete_text_before_caret(self, length=1):
         return NotImplementedError()
 
     def get_context(self):
@@ -104,14 +108,21 @@ class AtspiTextContext(TextContext):
         self._context = ""
         self._line = ""
         self._line_caret = 0
-        self._span_at_caret = TextSpan()
+        self._selection_span = TextSpan()
         self._begin_of_text = False        # context starts at begin of text?
         self._begin_of_text_offset = None  # offset of text begin
 
+        self._pending_separator_span = None
+        self._last_text_change_time = 0
+        self._last_caret_move_time = 0
+        self._last_caret_move_position = 0
+
         self._last_context = None
         self._last_line = None
 
         self._update_context_timer = Timer()
+        self._update_context_delay_normal = 0.01
+        self._update_context_delay = self._update_context_delay_normal
 
     def cleanup(self):
         self._register_atspi_listeners(False)
@@ -122,6 +133,15 @@ class AtspiTextContext(TextContext):
     def get_text_domain(self):
         return self._text_domain
 
+    def set_pending_separator(self, separator_span=None):
+        """ Remember this separator span for later insertion. """
+        if self._pending_separator_span is not separator_span:
+            self._pending_separator_span = separator_span
+
+    def get_pending_separator(self):
+        """ Return current pending separator span or None """
+        return self._pending_separator_span
+
     def get_context(self):
         """
         Returns the predictions context, i.e. some range of
@@ -155,6 +175,15 @@ class AtspiTextContext(TextContext):
 
         return context
 
+    def get_pending_bot_context(self):
+        """
+        Context including bot marker and pending separator.
+        """
+        context = self.get_bot_context()
+        if self._pending_separator_span is not None:
+            context += self._pending_separator_span.get_span_text()
+        return context
+
     def get_line(self):
         return self._line \
                if self._accessible else ""
@@ -167,13 +196,29 @@ class AtspiTextContext(TextContext):
         return self._line[self._line_caret:] \
                if self._accessible else ""
 
-    def get_span_at_caret(self):
-        return self._span_at_caret \
+    def get_selection_span(self):
+        return self._selection_span \
                if self._accessible else None
 
+    def get_span_at_caret(self):
+        if not self._accessible:
+            return None
+
+        span = self._selection_span.copy()
+        span.length = 0
+        return span
+
     def get_caret(self):
-        return self._span_at_caret.begin() \
-               if self._accessible else 0
+        return self._selection_span.begin() \
+            if self._accessible else 0
+
+    def get_character_extents(self, offset):
+        accessible = self._accessible
+        if accessible:
+            return self._state_tracker.get_accessible_character_extents(
+                accessible, offset)
+        else:
+            return None
 
     def get_text_begin_marker(self):
         domain = self.get_text_domain()
@@ -193,6 +238,12 @@ class AtspiTextContext(TextContext):
             return domain.can_suggest_before_typing()
         return True
 
+    def can_auto_punctuate(self):
+        domain = self.get_text_domain()
+        if domain:
+            return domain.can_auto_punctuate(self._begin_of_text)
+        return False
+
     def get_begin_of_text_offset(self):
         return self._begin_of_text_offset \
                if self._accessible else None
@@ -214,11 +265,11 @@ class AtspiTextContext(TextContext):
         #return False # support for inserting is spotty: not in firefox, terminal
         return bool(self._accessible) and self._can_insert_text
 
-    def delete_text(self, offset, length = 1):
+    def delete_text(self, offset, length=1):
         """ Delete directly, without going through faking key presses. """
         self._accessible.delete_text(offset, offset + length)
 
-    def delete_text_before_caret(self, length = 1):
+    def delete_text_before_caret(self, length=1):
         """ Delete directly, without going through faking key presses. """
         offset = self._accessible.get_caret_offset()
         self.delete_text(offset - length, length)
@@ -247,20 +298,20 @@ class AtspiTextContext(TextContext):
         offset = self._accessible.get_caret_offset()
         self.insert_text(offset, text)
 
-    def _register_atspi_listeners(self, register = True):
+    def _register_atspi_listeners(self, register=True):
         st = self._state_tracker
         if register:
             st.connect("text-entry-activated", self._on_text_entry_activated)
             st.connect("text-changed", self._on_text_changed)
             st.connect("text-caret-moved", self._on_text_caret_moved)
-            #st.connect("key-pressed", self._on_atspi_key_pressed)
+            # st.connect("key-pressed", self._on_atspi_key_pressed)
         else:
             st.disconnect("text-entry-activated", self._on_text_entry_activated)
             st.disconnect("text-changed", self._on_text_changed)
             st.disconnect("text-caret-moved", self._on_text_caret_moved)
-            #st.disconnect("key-pressed", self._on_atspi_key_pressed)
+            # st.disconnect("key-pressed", self._on_atspi_key_pressed)
 
-    def get_accessible_capabilities(accessible, **kwargs):
+    def get_accessible_capabilities(self, accessible, **kwargs):
         can_insert_text = False
         attributes = kwargs.get("attributes", {})
         interfaces = kwargs.get("interfaces", [])
@@ -268,17 +319,17 @@ class AtspiTextContext(TextContext):
         if accessible:
 
             # Can insert text via Atspi?
-            # Advantages: - faster, no individual key presses
-            #             - full trouble-free insertion of all unicode characters
+            # Advantages:
+            # - faster, no individual key presses
+            # - trouble-free insertion of all unicode characters
             if "EditableText" in interfaces:
                 # Support for atspi text insertion is spotty.
                 # Firefox, LibreOffice Writer, gnome-terminal don't support it,
                 # even if they claim to implement the EditableText interface.
 
-                # Allow direct text insertion by gtk widgets
-                if attributes and \
-                   "toolkit" in attributes and attributes["toolkit"] == "gtk":
-                   can_insert_text = True
+                # Allow direct text insertion for gtk widgets
+                if self._state_tracker.is_toolkit_gtk3(attributes):
+                    can_insert_text = True
 
         return can_insert_text
 
@@ -286,7 +337,6 @@ class AtspiTextContext(TextContext):
         # old text_domain still valid here
         self._wp.on_text_entry_deactivated()
 
-        #print("_on_text_entry_activated", accessible)
         # keep track of the active accessible asynchronously
         self._accessible = accessible
         self._entering_text = False
@@ -299,28 +349,33 @@ class AtspiTextContext(TextContext):
         self._text_domain.init_domain()
 
         # determine capabilities of this accessible
-        self._can_insert_text = self.get_accessible_capabilities(**state)
+        self._can_insert_text = \
+            self.get_accessible_capabilities(accessible, **state)
 
         # log accessible info
-        if _logger.isEnabledFor(logging.DEBUG):
-            log = _logger.debug
-            log("-"*70)
+        if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
+            log = _logger.atspi
+            log("-" * 70)
             log("Accessible focused: ")
+            indent = " " * 4
             if self._accessible:
                 state = self._state_tracker.get_state()
                 for key, value in sorted(state.items()):
                     msg = str(key) + "="
                     if key == "state-set":
                         msg += repr(AtspiStateType.to_strings(value))
+                    elif hasattr(value, "value_name"):  # e.g. role
+                        msg += value.value_name
                     else:
-                        msg += str(value)
-                    log(msg)
-                log("text_domain: {}".format(self._text_domain))
-                log("can_insert_text: {}".format(self._can_insert_text))
-                log("")
+                        msg += repr(value)
+                    log(indent + msg)
+                log(indent + "text_domain: {}"
+                    .format(self._text_domain and
+                            type(self._text_domain).__name__))
+                log(indent + "can_insert_text: {}"
+                    .format(self._can_insert_text))
             else:
-                log("None")
-                log("")
+                log(indent + "None")
 
         self._update_context()
 
@@ -330,26 +385,30 @@ class AtspiTextContext(TextContext):
         insertion_span = self._record_text_change(event.pos,
                                                   event.length,
                                                   event.insert)
-        # synchrounously notify of text insertion
+        # synchronously notify of text insertion
         if insertion_span:
             try:
                 caret_offset = self._accessible.get_caret_offset()
-            except: # gi._glib.GError
+            except:  # gi._glib.GError
                 pass
             else:
                 self._wp.on_text_inserted(insertion_span, caret_offset)
 
+        self._last_text_change_time = time.time()
         self._update_context()
 
     def _on_text_caret_moved(self, event):
+        self._last_caret_move_time = time.time()
+        self._last_caret_move_position = event.caret
         self._update_context()
+        self._wp.on_text_caret_moved()
 
     def _on_atspi_key_pressed(self, event):
         """ disabled, Francesco didn't receive any AT-SPI key-strokes. """
-        keycode = event.hw_code # uh oh, only keycodes...
-                                # hopefully "c" doesn't move around a lot.
-        modifiers = event.modifiers
-        #self._handle_key_press(keycode, modifiers)
+        # keycode = event.hw_code # uh oh, only keycodes...
+        #                         # hopefully "c" doesn't move around a lot.
+        # modifiers = event.modifiers
+        # self._handle_key_press(keycode, modifiers)
 
     def on_onboard_typing(self, key, mod_mask):
         if key.is_text_changing():
@@ -368,11 +427,11 @@ class AtspiTextContext(TextContext):
             domain = self.get_text_domain()
             if domain:
                 self._entering_text, end_of_editing = \
-                        domain.handle_key_press(keycode, modifiers)
+                    domain.handle_key_press(keycode, modifiers)
 
-                if end_of_editing == True:
+                if end_of_editing is True:
                     self._wp.commit_changes()
-                elif end_of_editing == False:
+                elif end_of_editing is False:
                     self._wp.discard_changes()
 
     def _record_text_change(self, pos, length, insert):
@@ -383,17 +442,15 @@ class AtspiTextContext(TextContext):
         if accessible:
             try:
                 char_count = accessible.get_character_count()
-            except: # gi._glib.GError: The application no longer exists
-                    # when closing a tab in gnome-terminal.
+            except:     # gi._glib.GError: The application no longer exists
+                        # when closing a tab in gnome-terminal.
                 char_count = None
 
-        if not char_count is None:
+        if char_count is not None:
             # record the change
             spans_to_update = []
 
             if insert:
-                #print("insert", pos, length)
-
                 if self._entering_text and \
                    self.can_record_insertion(accessible, pos, length):
                     if self._wp.is_typing() or length < 30:
@@ -408,10 +465,10 @@ class AtspiTextContext(TextContext):
 
                     # simple span for current insertion
                     begin = max(pos - 100, 0)
-                    end = min(pos+length + 100, char_count)
+                    end = min(pos + length + 100, char_count)
                     text = self._state_tracker.get_accessible_text(accessible,
                                                                    begin, end)
-                    if not text is None:
+                    if text is not None:
                         insertion_span = TextSpan(pos, length, text, begin)
                 else:
                     # Remember nothing, just update existing spans.
@@ -421,7 +478,6 @@ class AtspiTextContext(TextContext):
                                                        include_length)
 
             else:
-                #print("delete", pos, length)
                 spans_to_update = self._changes.delete(pos, length,
                                                        self._entering_text)
 
@@ -434,31 +490,49 @@ class AtspiTextContext(TextContext):
                 span.text = Atspi.Text.get_text(accessible, begin, end)
                 span.text_pos = begin
 
-            #print(self._changes)
-
         self._text_changed = True
 
         return insertion_span
 
+    def set_update_context_delay(self, delay):
+        self._update_context_delay = delay
+
+    def reset_update_context_delay(self):
+        self._update_context_delay = self._update_context_delay_normal
+
     def _update_context(self):
-        self._update_context_timer.start(0.01, self.on_text_context_changed)
+        self._update_context_timer.start(self._update_context_delay,
+                                         self.on_text_context_changed)
 
     def on_text_context_changed(self):
+        # Clear pending separator when the user clicked to move
+        # the cursor away from the separator position.
+        if self._pending_separator_span:
+            # Lone caret movement, no recent text change?
+            if self._last_caret_move_time - self._last_text_change_time > 1.0:
+                # Away from the separator?
+                if self._last_caret_move_position != \
+                   self._pending_separator_span.begin():
+                    self.set_pending_separator(None)
+
         result = self._text_domain.read_context(self._accessible)
-        if not result is None:
+        if result is not None:
             (self._context,
              self._line,
              self._line_caret,
-             self._span_at_caret,
+             self._selection_span,
              self._begin_of_text,
              self._begin_of_text_offset) = result
 
-            context = self.get_bot_context() # make sure to include bot-markers
-            if self._last_context != context or \
-               self._last_line != self._line:
+            # make sure to include bot-markers and pending separator
+            context = self.get_pending_bot_context()
+            change_detected = (self._last_context != context or
+                               self._last_line != self._line)
+            if change_detected:
                 self._last_context = context
                 self._last_line    = self._line
-                self._wp.on_text_context_changed()
+
+            self._wp.on_text_context_changed(change_detected)
 
         return False
 
@@ -493,12 +567,12 @@ class InputLine(TextContext):
         self.line   = self.line[:self.caret] + s + self.line[self.caret:]
         self.move_caret(len(s))
 
-    def delete_left(self, n=1):  # backspace
-        self.line = self.line[:self.caret-n] + self.line[self.caret:]
+    def delete_left(self, n=1):     # backspace
+        self.line = self.line[:self.caret - n] + self.line[self.caret:]
         self.move_caret(-n)
 
-    def delete_right(self, n=1): # delete
-        self.line = self.line[:self.caret] + self.line[self.caret+n:]
+    def delete_right(self, n=1):    # delete
+        self.line = self.line[:self.caret] + self.line[self.caret + n:]
 
     def move_caret(self, n):
         self.caret += n
@@ -527,8 +601,9 @@ class InputLine(TextContext):
         """
         if char == "\t":
             return True
-        return not unicodedata.category(char) in ('Cc','Cf','Cs','Co',
-                                                  'Cn','Zl','Zp')
+        return not unicodedata.category(char) in ('Cc', 'Cf', 'Cs', 'Co',
+                                                  'Cn', 'Zl', 'Zp')
+
     def track_sent_key(self, key, mods):
         """
         Sync input_line with single key presses.
@@ -537,16 +612,15 @@ class InputLine(TextContext):
         end_editing = False
 
         if config.wp.stealth_mode:
-            return  True
+            return True
 
         id = key.id.upper()
         char = key.get_label()
-        #print  id," '"+char +"'",key.action_type
         if char is None or len(char) > 1:
             char = ""
 
         if key.action_type == KeyCommon.WORD_ACTION:
-            pass # don't reset input on word insertion
+            pass  # don't reset input on word insertion
 
         elif key.action_type == KeyCommon.MODIFIER_ACTION:
             pass  # simply pressing a modifier shouldn't stop the word
@@ -555,12 +629,12 @@ class InputLine(TextContext):
             pass
 
         elif key.action_type == KeyCommon.KEYSYM_ACTION:
-            if   id == 'ESC':
+            if id == 'ESC':
                 self.reset()
             end_editing = True
 
         elif key.action_type == KeyCommon.KEYPRESS_NAME_ACTION:
-            if   id == 'DELE':
+            if id == 'DELE':
                 self.delete_right()
             elif id == 'LEFT':
                 self.move_caret(-1)
@@ -570,7 +644,7 @@ class InputLine(TextContext):
                 end_editing = True
 
         elif key.action_type == KeyCommon.KEYCODE_ACTION:
-            if   id == 'RTRN':
+            if id == 'RTRN':
                 char = "\n"
             elif id == 'SPCE':
                 char = " "
@@ -589,9 +663,9 @@ class InputLine(TextContext):
         else:
             end_editing = True
 
-        if not self.is_valid(): # caret moved outside known range?
+        if not self.is_valid():  # caret moved outside known range?
             end_editing = True
 
-        #print end_editing,"'%s' " % self.line, self.caret
+        # print end_editing,"'%s' " % self.line, self.caret
         return end_editing
 
