@@ -1,4 +1,9 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2015 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
 # Onboard is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 3 of the License, or
@@ -11,10 +16,6 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
-#
-# Copyright © 2012, marmuta
-#
-# This file is part of Onboard.
 
 from __future__ import division, print_function, unicode_literals
 
@@ -23,8 +24,9 @@ import re
 import glob
 
 import logging
-_logger = logging.getLogger("TextDomain")
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 try:
     from gi.repository import Atspi
 except ImportError as e:
@@ -33,6 +35,7 @@ except ImportError as e:
 from Onboard.TextChanges  import TextSpan
 from Onboard.utils        import KeyCode, Modifiers, unicode_str
 
+_logger = logging.getLogger("TextDomain")
 
 class TextDomains:
     """ Collection of all recognized text domains. """
@@ -58,7 +61,9 @@ class TextDomains:
 
 
 class TextDomain:
-    """ Abstract base class as a catch-all for domain specific functionalty. """
+    """
+    Abstract base class as a catch-all for domain specific functionalty.
+    """
 
     def __init__(self):
         self._url_parser = PartialURLParser()
@@ -98,17 +103,25 @@ class TextDomain:
         ''
 
         # filename
+        >>> import tempfile
+        >>> from os.path import abspath, dirname, join
+        >>> def touch(fn):
+        ...     with open(fn, mode="w") as f: n = f.write("")
+        >>> td = tempfile.TemporaryDirectory(prefix="test_onboard _")
+        >>> dir = td.name
+
+        >>> touch(join(dir, "onboard-defaults.conf.example"))
+        >>> touch(join(dir, "onboard-defaults.conf.example.another"))
+        >>>
         >>> d.get_auto_separator("/etc")
         '/'
-        >>> from os.path import abspath, dirname
-        >>> path = dirname(abspath("./onboard-defaults.conf.example"))
-        >>> d.get_auto_separator(join(path, "onboard-defaults"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults"))
         '.'
-        >>> d.get_auto_separator(join(path, "onboard-defaults.conf"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults.conf"))
         '.'
-        >>> d.get_auto_separator(join(path, "onboard-defaults.conf.example"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults.conf.example"))
         ''
-        >>> d.get_auto_separator(join(path, "onboard-defaults.conf.example.nexus7"))
+        >>> d.get_auto_separator(join(dir, "onboard-defaults.conf.example.another"))
         ' '
 
         # filename in directory with spaces
@@ -116,12 +129,12 @@ class TextDomain:
         >>> td = tempfile.TemporaryDirectory(prefix="test onboard _")
         >>> dir = td.name
         >>> fn = os.path.join(dir, "onboard-defaults.conf.example")
-        >>> with open(fn, mode="w") as f: n = f.write("")
+        >>> touch(fn)
         >>> d.get_auto_separator(join(dir, "onboard-defaults"))
         '.'
 
         # no false positives for valid filenames before the current token
-        >>> d.get_auto_separator(path + "/onboard-defaults no-file")
+        >>> d.get_auto_separator(dir + "/onboard-defaults no-file")
         ' '
         """
         separator = " "
@@ -203,19 +216,19 @@ class TextDomain:
         Doctests:
         >>> d = DomainGenericText()
 
-        # Span doesn't change for simple words
+        # Span doesn't grow for simple words
         >>> d.grow_learning_span(TextSpan(8, 1, "word1 word2 word3"))
         (8, 1, 'r')
 
-        # Span growing to cover a complete URL
+        # Span grows to include a complete URL
         >>> d.grow_learning_span(TextSpan(13, 1, "http://www.domain.org"))
         (0, 21, 'http://www.domain.org')
 
-        # Span growing to cover multiple complete URLs
+        # Span grows to include multiple complete URLs
         >>> d.grow_learning_span(TextSpan(19, 13, "http://www.domain.org word http://slashdot.org"))
         (0, 46, 'http://www.domain.org word http://slashdot.org')
 
-        # Span growing to cover a complete filename
+        # Span grows to include a complete filename
         >>> d.grow_learning_span(TextSpan(10, 1, "word1 /usr/bin/bash word2"))
         (6, 13, '/usr/bin/bash')
 
@@ -229,8 +242,9 @@ class TextDomain:
         >>> d.grow_learning_span(TextSpan(18, 1, "word1 /usr/bin/bash word2"))
         (6, 13, '/usr/bin/bash')
 
-        # Large text with text offset > 0: returned position must be offset too.
-        >>> d.grow_learning_span(TextSpan(116, 1, "word1 /usr/bin/bash word2", 100))
+        # Large text with text offset>0: returned position must be offset too
+        >>> d.grow_learning_span(TextSpan(116, 1,
+        ...     "word1 /usr/bin/bash word2", 100))
         (106, 13, '/usr/bin/bash')
         """
         text   = text_span.get_text()
@@ -238,30 +252,32 @@ class TextDomain:
         begin  = text_span.begin() - offset
         end    = text_span.end() - offset
 
-        tokens, spans = self._split_growth_sections(text)
+        sections, spans = self._split_growth_sections(text)
 
         for i, s in enumerate(spans):
-            if begin < s[1] and end > s[0]: # intersects?
+            if begin < s[1] and end > s[0]:  # intersects?
 
-                token = tokens[i]
+                section = sections[i]
                 span = spans[i]
-                if self._url_parser.is_maybe_url(token) or \
-                   self._url_parser._is_maybe_filename(token):
+                if self._url_parser.is_maybe_url(section) or \
+                   self._url_parser._is_maybe_filename(section):
                     begin = min(begin, span[0])
                     end = max(end, span[1])
 
-        return begin+offset, end - begin, text[begin:end]
-
+        return begin + offset, end - begin, text[begin:end]
     def can_record_insertion(self, accessible, pos, length):
         return True
 
     def can_give_keypress_feedback(self):
         return True
 
-    def can_spell_check(self):
+    def can_spell_check(self, section_span):
         return False
 
-    def can_auto_correct(self):
+    def can_auto_correct(self, section_span):
+        return False
+
+    def can_auto_punctuate(self, has_begin_of_text):
         return False
 
     def can_suggest_before_typing(self):
@@ -271,7 +287,7 @@ class TextDomain:
     def handle_key_press(self, keycode, mod_mask):
         return True, None  # entering_text, end_of_editing
 
-    _growth_tokens_pattern = re.compile("[^\s?#@]+", re.DOTALL)
+    _growth_sections_pattern = re.compile("[^\s?#@]+", re.DOTALL)
 
     def _split_growth_sections(self, text):
         """
@@ -289,7 +305,7 @@ class TextDomain:
         >>> d._split_growth_sections("http://user:pass@www.domain.org")
         (['http://user:pass', 'www.domain.org'], [(0, 16), (17, 31)])
         """
-        matches = self._growth_tokens_pattern.finditer(text)
+        matches = self._growth_sections_pattern.finditer(text)
         tokens = []
         spans = []
         for m in matches:
@@ -368,7 +384,13 @@ class DomainGenericText(TextDomain):
 
         begin = max(selection[0] - 256, 0)
         end   = min(selection[0] + 100, count)
-        text = Atspi.Text.get_text(accessible, begin, end)
+        try:
+            text = Atspi.Text.get_text(accessible, begin, end)
+        except Exception as ex: # Private exception gi._glib.GError when
+                                # gedit became unresponsive.
+            _logger.info("DomainGenericText.read_context(), text2: " \
+                         + unicode_str(ex))
+            return None
 
         text = unicode_str(text)
 
@@ -383,10 +405,31 @@ class DomainGenericText(TextDomain):
         return (context, line, line_caret, selection_span,
                 begin_of_text, begin_of_text_offset)
 
-    def can_spell_check(self):
+    def can_spell_check(self, section_span):
+        """
+        Can we auto-correct this span?.
+
+        Doctests:
+        >>> d = DomainGenericText()
+        >>> d.can_spell_check(TextSpan(0, 3, "abc"))
+        True
+        >>> d.can_spell_check(TextSpan(4, 3, "abc def"))
+        True
+        >>> d.can_spell_check(TextSpan(0, 20, "http://www.domain.org"))
+        False
+        """
+        section = section_span.get_span_text()
+        return not self._url_parser.is_maybe_url(section) and \
+               not self._url_parser._is_maybe_filename(section)
         return True
 
-    def can_auto_correct(self):
+    def can_auto_correct(self, section_span):
+        """
+        Can we auto-correct this span?.
+        """
+        return self.can_spell_check(section_span)
+
+    def can_auto_punctuate(self, has_begin_of_text):
         return True
 
     def get_text_begin_marker(self):
@@ -396,22 +439,24 @@ class DomainGenericText(TextDomain):
 class DomainTerminal(TextDomain):
     """ Terminal entry, in particular gnome-terminal """
 
-    _prompt_patterns = tuple(re.compile(p, re.UNICODE) for p in \
-                             ("^gdb$ ",
-                              "^>>> ", # python
-                              "^In \[[0-9]*\]: ",   # ipython
-                              "^:",    # vi command mode
-                              "^/",    # vi search
-                              "^\?",   # vi reverse search
-                              "\$ ",   # generic prompt
-                              "# ",    # root prompt
-                             )
+    _prompt_patterns = tuple(re.compile(p, re.UNICODE) for p in
+                                (
+                                    "^gdb$ ",
+                                    "^>>> ",              # python
+                                    "^In \[[0-9]*\]: ",   # ipython
+                                    "^:",                 # vi command mode
+                                    "^/",                 # vi search
+                                    "^\?",                # vi reverse search
+                                    "\$ ",                # generic prompt
+                                    "# ",                 # root prompt
+                                    "^.*?@.*?/.*?> "      # fish
+                                )
                             )
 
-    _prompt_blacklist_patterns = tuple(re.compile(p, re.UNICODE) for p in \
-                             (
-                              "^\(.*\)`.*': ", # bash incremental search
-                             )
+    _prompt_blacklist_patterns = tuple(re.compile(p, re.UNICODE) for p in
+                                (
+                                    "^\(.*\)`.*': ",  # bash incremental search
+                                )
                             )
 
     def matches(self, **kwargs):
@@ -421,21 +466,21 @@ class DomainTerminal(TextDomain):
     def init_domain(self):
         pass
 
-    def read_context(self, accessible, offset = None):
+    def read_context(self, accessible, offset=None):
         """
         Extract prediction context from the accessible
         """
         if offset is None:
             try:
                 offset = accessible.get_caret_offset()
-            except Exception as ex: # Private exception gi._glib.GError when
-                                    # gedit became unresponsive.
-                _logger.info("DomainTerminal.read_context(): " \
-                             + unicode_str(ex))
+            except Exception as ex:     # Private exception gi._glib.GError
+                                        # when gedit became unresponsive.
+                _logger.info("DomainTerminal.read_context(): " +
+                             unicode_str(ex))
                 return None
 
         context_lines, prompt_length, line, line_start, line_caret = \
-                             self._get_text_after_prompt(accessible, offset)
+            self._get_text_after_prompt(accessible, offset)
 
         if prompt_length:
             begin_of_text = True
@@ -448,7 +493,7 @@ class DomainTerminal(TextDomain):
         before_line = "".join(context_lines[:-1])
         selection_span = TextSpan(offset, 0,
                                   before_line + line,
-                                  line_start-len(before_line))
+                                  line_start - len(before_line))
 
         result = (context, line, line_caret, selection_span,
                   begin_of_text, begin_of_text_offset)
@@ -527,7 +572,7 @@ class DomainTerminal(TextDomain):
             prompt_length = self._find_prompt(l)
             context_lines.insert(0, l[prompt_length:])
             if i == 0:
-                line = line[prompt_length:] # cut prompt from input line
+                line = line[prompt_length:]  # cut prompt from input line
                 line_start += prompt_length
                 line_caret -= prompt_length
             if prompt_length:
@@ -569,7 +614,7 @@ class DomainTerminal(TextDomain):
         # Problem: learning won't happen for uncommon prompts, but less random
         # junk scrolling by should enter the user model in return.
         context_lines, prompt_length, line, line_start, line_caret = \
-                self._get_text_after_prompt(accessible, offset)
+            self._get_text_after_prompt(accessible, offset)
         return bool(prompt_length)
 
     def can_suggest_before_typing(self):
@@ -591,6 +636,14 @@ class DomainTerminal(TextDomain):
 
         return True, None  # entering_text, end_of_editing
 
+    def can_auto_punctuate(self, has_begin_of_text):
+        """
+        Only auto-punctuate in Terminal when no prompt was detected.
+        Intention is to allow punctuation assistance in editors, but disable
+        it when entering commands at the prompt, e.g. for "cd ..".
+        """
+        return not has_begin_of_text
+
 
 class DomainURL(DomainGenericText):
     """ (Firefox) address bar """
@@ -612,7 +665,7 @@ class DomainURL(DomainGenericText):
     def get_text_begin_marker(self):
         return "<bot:url>"
 
-    def can_spell_check(self):
+    def can_spell_check(self, section_span):
         return False
 
 
@@ -751,10 +804,14 @@ class PartialURLParser:
         '.'
         >>> p.get_auto_separator("http://www.domain.org")
         '/'
-        >>> p.get_auto_separator("http://www.domain.co")
-        '.'
+        >>> p.get_auto_separator("http://www.domain.co") # ambiguous co/ or co.uk/
+        '/'
         >>> p.get_auto_separator("http://www.domain.co.uk")
         '/'
+        >>> p.get_auto_separator("http://www.domain.co.uk/home")
+        '/'
+        >>> p.get_auto_separator("http://www.domain.co/home")
+        '/'
         >>> p.get_auto_separator("http://www.domain.org/home")
         '/'
         >>> p.get_auto_separator("http://www.domain.org/home/index.html")
@@ -763,26 +820,36 @@ class PartialURLParser:
         ':'
 
         # local files
+        >>> import tempfile
+        >>> from os.path import abspath, dirname, join
+        >>> def touch(fn):
+        ...     with open(fn, mode="w") as f: n = f.write("")
+        >>> td = tempfile.TemporaryDirectory(prefix="test onboard _")
+        >>> dir = td.name
+
+        >>> touch(join(dir, "onboard-defaults.conf.example"))
+        >>> touch(join(dir, "onboard-defaults.conf.example.another"))
+        >>>
+        >>> import glob
+        >>> #glob.glob(dir+"/**")
         >>> p.get_auto_separator("file")
         ':///'
         >>> p.get_auto_separator("file:///home")
         '/'
-        >>> from os.path import abspath, dirname
-        >>> url = "file://" + dirname(abspath("./onboard-defaults.conf.example"))
-        >>> p.get_auto_separator(url + "/onboard-defaults")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults"))
         '.'
-        >>> p.get_auto_separator(url + "/onboard-defaults.conf")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults.conf"))
         '.'
-        >>> p.get_auto_separator(url + "/onboard-defaults.conf.example")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults.conf.example"))
         ''
-        >>> p.get_auto_separator(url + "/onboard-defaults.conf.example.nexus7")
+        >>> p.get_auto_separator("file://"+join(dir, "onboard-defaults.conf.example.another"))
         ' '
 
         # Non-existing filename: we don't know, don't guess a separator
         >>> p.get_auto_separator("file:///tmp/onboard1234")
         ''
 
-        # Non-existing filename: if basename has an extension assumes we're done
+        # Non-existing filename: if basename has an extension assume we're done
         >>> p.get_auto_separator("file:///tmp/onboard1234.txt")
         ' '
 
@@ -796,14 +863,18 @@ class PartialURLParser:
         SCHEME, PROTOCOL, DOMAIN, PATH = range(4)
         component = SCHEME
         last_septok = ""
-        matches = self.iter_url(context)
-        for match in matches:
+        matches = tuple(self.iter_url(context))
+        for index, match in enumerate(matches):
             groups = match.groups()
             token  = groups[0]
             septok = groups[1]
 
             if septok:
                 last_septok = septok
+            if index < len(matches)-1:
+                next_septok = matches[index+1].groups()[1]
+            else:
+                next_septok = ""
 
             if component == SCHEME:
                 if token:
@@ -826,8 +897,8 @@ class PartialURLParser:
                 if token:
                     separator = "."
                     if last_septok == "." and \
-                       token in self._TLDs and \
-                       token != "co":  # special case for co.uk TLD
+                       next_septok != "." and \
+                       token in self._TLDs:
                         separator = "/"
                         component = PATH
                         continue
@@ -860,7 +931,7 @@ class PartialURLParser:
 
         if os.path.isabs(filename):
             files  = glob.glob(filename + "*")
-            files += glob.glob(filename + "/*") # look inside directories too
+            files += glob.glob(filename + "/*")  # look inside directories too
             separator = self._get_separator_from_file_list(filename, files)
 
         if separator is None:
@@ -913,6 +984,14 @@ class PartialURLParser:
         '.'
         >>> p._get_separator_from_file_list("/dir/dir/file.ext1.ext2", files)
         ' '
+
+        # partial path match
+        >>> files = ["/dir/dir/file", "/dir/dir/file.ext1",
+        ...          "/dir/dir/file.ext2"]
+        >>> p._get_separator_from_file_list("/dir/dir/file", files)
+        ''
+        >>> p._get_separator_from_file_list("/dir/dir/file.ext1", files)
+        ' '
         """
         separator = None
 
