@@ -1,9 +1,27 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2014 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 """ Enlarged drag handles for resizing or moving """
 
 from __future__ import division, print_function, unicode_literals
 
-from math import pi, sqrt, sin, log
+from math import pi, sqrt
 import cairo
 
 from Onboard.utils       import Rect, drop_shadow
@@ -11,7 +29,6 @@ from Onboard.definitions import Handle
 
 ### Logging ###
 import logging
-from functools import reduce
 _logger = logging.getLogger("TouchHandles")
 ###############
 
