@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright © 2013, marmuta
+# Copyright © 2013-2014 marmuta <marmvta@gmail.com>
 #
 # This file is part of Onboard.
 #
@@ -21,15 +21,16 @@ from __future__ import division, print_function, unicode_literals
 
 import os
 import time
+import logging
+
+from Onboard.utils import unicode_str, XDGDirs
+from Onboard.Timer import Timer
+from Onboard.Config import Config
 
 import Onboard.pypredict as pypredict
 
-import Onboard.utils as utils
-
-from Onboard.Config import Config
 config = Config()
 
-import logging
 _logger = logging.getLogger(__name__)
 
 
@@ -83,13 +84,24 @@ class WPLocalEngine(object):
         """
         self._model_cache.get_models(self.models)
 
-    def predict(self, context_line, limit = 20,
-                case_insensitive = False,
-                case_insensitive_smart = False,
-                accent_insensitive = False,
-                accent_insensitive_smart = False,
-                ignore_capitalized = False,
-                ignore_non_capitalized = False):
+    def postpone_autosave(self):
+        self._auto_save_timer.postpone()
+
+    def pause_autosave(self):
+        # Pause for a minute max, because resume_autosave isn't
+        # reliable called, e.g. when dragging and leaving the window.
+        self._auto_save_timer.pause(60)
+
+    def resume_autosave(self):
+        self._auto_save_timer.resume()
+
+    def predict(self, context_line, limit=20,
+                case_insensitive=False,
+                case_insensitive_smart=False,
+                accent_insensitive=False,
+                accent_insensitive_smart=False,
+                ignore_capitalized=False,
+                ignore_non_capitalized=False):
         """ Find completion/prediction choices. """
         LanguageModel = pypredict.LanguageModel
         options = 0
@@ -117,20 +129,29 @@ class WPLocalEngine(object):
         if self.auto_learn_models:
             tokens, spans = pypredict.tokenize_text(text)
 
-            # Remove trailing single quote, too many false positives.
+            # There are too many false positives with trailing
+            # single quotes, remove them.
             # Do this here, because we still want "it's", etc. to
             # incrementally provide completions.
             for i, token in enumerate(tokens):
                 if token.endswith("'"):
                     token = token[:-1]
-                    if not token: # shouldn't happen
+                    if not token:  # shouldn't happen
                         token = "<unk>"
                     tokens[i] = token
 
+            # if requested, drop unknown words
+            if allow_new_words:
+                token_sections = [tokens]
+            else:
+                token_sections = self._drop_new_words(tokens, spans,
+                                                      self.persistent_models)
             models = self._model_cache.get_models(self.auto_learn_models)
             for model in models:
-                model.learn_tokens(tokens, allow_new_words)
-            _logger.info("learn_text: tokens=" + repr(tokens[:10]))
+                for tokens in token_sections:
+                    model.learn_tokens(tokens)
+
+            _logger.info("learn_text: tokens=" + repr(token_sections))
 
             # debug: save all learned text for later parameter optimization
             if config.log_learn:
@@ -138,12 +159,20 @@ class WPLocalEngine(object):
                 with open(fn, "a") as f:
                     f.write(text + "\n")
 
+    def _drop_new_words(self, tokens, spans, lmids):
+        """ Remove tokens that don't already exist in any active model.  """
+
+        tokspans, counts = self.lookup_tokens(tokens, spans, lmids)
+        split_indices = [i for i, model_counts in enumerate(counts)
+                         if all(n != 1 for n in model_counts)]
+        return pypredict.split_tokens_at(tokens, split_indices)
+
     def learn_scratch_text(self, text):
         """ Count n-grams and add words to the scratch models. """
         tokens, spans = pypredict.tokenize_text(text)
         models = self._model_cache.get_models(self.scratch_models)
         for model in models:
-            #print("scratch learn", model, tokens)
+            # print("scratch learn", model, tokens)
             model.learn_tokens(tokens, True)
 
     def clear_scratch_models(self):
@@ -152,12 +181,19 @@ class WPLocalEngine(object):
         for model in models:
             model.clear()
 
-    def lookup_text(self, text):
+    def lookup_text(self, text, lmids):
         """
         Split <text> into tokens and lookup the individual tokens in each
-        of the given language models. This method is meant to be a basis for
-        highlighting (partially) unknown words in a display for recently
-        typed text.
+        of the given language models. See lookup_tokens() for more information.
+        """
+        tokens, spans = pypredict.tokenize_sentence(text)
+        return self.lookup_tokens(tokens, spans, lmids)
+
+    def lookup_tokens(self, tokens, spans, lmids):
+        """
+        Lookup the individual tokens in each of the given language models.
+        This method is meant to be a basis for highlighting (partially)
+        unknown words in a display for recently typed text.
 
         The return value is a tuple of two arrays. First an array of tuples
         (start, end, token), one per token, with start and end index pointing
@@ -166,20 +202,20 @@ class WPLocalEngine(object):
         result is either 0 for no match, 1 for an exact match or -n for
         count n partial (prefix) matches.
         """
-        lmids = self.models
-        toks, spans = pypredict.tokenize_sentence(text)
-        tokens  = [(spans[i][0], spans[i][1], t) for i,t in enumerate(toks)]
-        counts = [[0 for lmid in lmids] for t in tokens]
-        for i,lmid in enumerate(lmids):
+        tokspans  = [(spans[i][0], spans[i][1], t)
+                     for i, t in enumerate(tokens)]
+        counts = [[0 for lmid in lmids] for t in tokspans]
+        for i, lmid in enumerate(lmids):
             model = self._model_cache.get_model(lmid)
             if model:
-                for j,t in enumerate(tokens):
+                for j, t in enumerate(tokspans):
                     counts[j][i] = model.lookup_word(t[2])
 
-        _logger.debug("lookup_words: tokens=%s counts=%s" % \
-                     (repr(tokens), repr(counts)) )
+        _logger.debug("lookup_tokens: tokens=%s counts=%s" %
+                     (repr(tokens), repr(counts)))
 
-        # counts are 0 for no match, 1 for exact match or -n for partial matches
+        # Counts are 0 for no match, 1 for exact match or
+        # -n for partial matches
         return tokens, counts
 
     def word_exists(self, word):
@@ -188,7 +224,7 @@ class WPLocalEngine(object):
         """
         exists = False
         lmids = self.persistent_models
-        for i,lmid in enumerate(lmids):
+        for i, lmid in enumerate(lmids):
             model = self._model_cache.get_model(lmid)
             if model:
                 count = model.lookup_word(word)
@@ -233,7 +269,7 @@ class WPLocalEngine(object):
         text = text[-1024:]
         tokens, spans = self.tokenize_context(text)
         if len(spans):
-            # Don't return the the token itself as it won't include
+            # Don't return the token itself as it won't include
             # trailing dashes. Catch the text until its very end.
             begin = spans[-1][0]
             return text[begin:]
@@ -249,7 +285,7 @@ class WPLocalEngine(object):
             # failed in practice for anything but natural language, e.g.
             # shell commands.
             # -> use the second best available: absolute discounting
-            #m.smoothing = "kneser-ney"
+            # m.smoothing = "kneser-ney"
             m.smoothing = "abs-disc"
 
             # setup recency caching
@@ -266,13 +302,33 @@ class WPLocalEngine(object):
                 m.recency_lambdas = [0.404, 0.831, 0.444]
 
         model = pypredict.overlay(models)
-        #model = pypredict.linint(models, weights)
-        #model = pypredict.loglinint(models, weights)
+        # model = pypredict.linint(models, weights)
+        # model = pypredict.loglinint(models, weights)
 
         choices = model.predictp(context, limit, options=options)
 
         return choices
 
+    def remove_context(self, context):
+        """
+        Remove the last word of context in the given context.
+        If len(context) == 1 then all occurences of the word will be removed.
+        """
+        lmids, weights = self._model_cache.parse_lmdesc(self.auto_learn_models)
+        models = self._model_cache.get_models(lmids)
+        for i, m in enumerate(models):
+            changes = m.remove_context(context)
+
+            # debug output
+            _logger.debug("removing {} from '{}': {} n-grams affected"
+                          .format(context, lmids[i], len(changes)))
+            if _logger.isEnabledFor(logging.DEBUG):
+                changes = sorted(sorted(changes.items()),
+                                 key=lambda x: -len(x[0]))
+                for ng in changes:
+                    _logger.debug("    remove: {}, count {}"
+                                  .format(ng[0], ng[1]))
+
 
 class ModelCache:
     """ Loads and caches language models """
@@ -327,7 +383,7 @@ class ModelCache:
                 if filename.endswith("." + extension):
                     models.append(os.path.join(path, filename))
         except OSError as e:
-            _logger.warning("Failed to find language models in '{}': {} ({})" \
+            _logger.warning("Failed to find language models in '{}': {} ({})"
                             .format(path, os.strerror(e.errno), e.errno))
         return models
 
@@ -346,7 +402,7 @@ class ModelCache:
             lmids.append(fields[0])
 
             weight = 1.0
-            if len(fields) >= 2: # weight is optional
+            if len(fields) >= 2:  # weight is optional
                 try:
                     weight = float(fields[1])
                 except:
@@ -384,7 +440,7 @@ class ModelCache:
         filename = self.get_filename(lmid)
 
         if type_ == "lm":
-            if   class_ == "system":
+            if class_ == "system":
                 if pypredict.read_order(filename) == 1:
                     model = pypredict.UnigramModel()
                 else:
@@ -394,11 +450,11 @@ class ModelCache:
             elif class_ == "mem":
                 model = pypredict.DynamicModel()
             else:
-                _logger.error("Unknown class component '{}' in lmid '{}'" \
+                _logger.error("Unknown class component '{}' in lmid '{}'"
                               .format(class_, lmid))
                 return None
         else:
-            _logger.error("Unknown type component '{}' in lmid '{}'" \
+            _logger.error("Unknown type component '{}' in lmid '{}'"
                           .format(type_, lmid))
             return None
 
@@ -414,20 +470,20 @@ class ModelCache:
         if not os.path.exists(filename):
             if class_ == "system":
                 _logger.warning("System language model '{}' "
-                                "doesn't exist, skipping." \
+                                "doesn't exist, skipping."
                                 .format(filename))
         else:
             try:
                 model.load(filename)
             except IOError as ex:
-                if not ex.errno is None: # not n-gram count mismatch
+                if ex.errno is not None:  # not n-gram count mismatch
                     errno = ex.errno
                     errstr = os.strerror(errno)
                     msg = _format(
-                            "Failed to load language model '{}': {} ({})", \
+                            "Failed to load language model '{}': {} ({})",
                             filename, errstr, errno)
                 else:
-                    msg = utils.unicode_str(ex)
+                    msg = unicode_str(ex)
                 _logger.error(msg)
                 model.load_error_msg = msg
 
@@ -448,6 +504,7 @@ class ModelCache:
     def save_model(self, model, lmid):
         type_, class_, name  = lmid.split(":")
         filename = self.get_filename(lmid)
+
         backup_filename = self.get_backup_filename(filename)
 
         if filename and \
@@ -455,14 +512,14 @@ class ModelCache:
 
             if model.load_error:
                 _logger.warning("Not saving modified language model '{}' "
-                                "due to previous error on load." \
+                                "due to previous error on load."
                                 .format(filename))
             else:
                 _logger.info("Saving language model '{}'".format(filename))
                 try:
                     # create the path
                     path = os.path.dirname(filename)
-                    utils.XDGDirs.assure_user_dir_exists(path)
+                    XDGDirs.assure_user_dir_exists(path)
 
                     if 1:
                         # save to temp file
@@ -477,8 +534,9 @@ class ModelCache:
 
                     model.modified = False
                 except (IOError, OSError) as e:
-                    _logger.warning("Failed to save language model '{}': {} ({})" \
-                                    .format(filename, os.strerror(e.errno), e.errno))
+                    _logger.warning(
+                        "Failed to save language model '{}': {} ({})"
+                        .format(filename, os.strerror(e.errno), e.errno))
 
     @staticmethod
     def get_filename(lmid):
@@ -488,7 +546,7 @@ class ModelCache:
         else:
             if class_ == "system":
                 path = config.get_system_model_dir()
-            else: #if class_ == "user":
+            else:  # if class_ == "user":
                 path = config.get_user_model_dir()
             ext = type_
             filename = os.path.join(path, name + "." + ext)
@@ -537,19 +595,65 @@ class ModelCache:
         return fn
 
 
-class AutoSaveTimer(utils.Timer):
+class AutoSaveTimer(Timer):
     """ Auto-save modified language models periodically """
 
-    def __init__(self, mode_cache, interval = 10*60):
+    def __init__(self, mode_cache,
+                 interval_min=10 * 60,
+                 interval_max=30 * 60,
+                 postpone_delay=10):
         self._model_cache = mode_cache
-        self._interval = interval  # in seconds
-        self._last_save_time = 0
-        self.start(5, self._on_timer)
+        self._interval_min = interval_min  # in seconds
+        self._interval_max = interval_max  # in seconds
+        self._postpone_delay = postpone_delay
+        self._interval = self._interval_min  # in seconds
+        self._last_save_time = time.time()
+        self._pause = 0
+        self._timer_interval = 5
+        self.start(self._timer_interval, self._on_timer)
+
+    def pause(self, duration=None):
+        """
+        No auto-saving while paused, e.g. during key-press.
+        """
+        self._pause = duration
+
+    def resume(self):
+        """
+        Allow auto-saving again.
+        """
+        self._pause = 0
+
+    def postpone(self):
+        """
+        Postpone saving a little, while the user is still typing.
+        Helps to mask the delay when saving large models during which
+        Onboard briefly becomes unresponsive.
+        """
+        elapsed = time.time() - self._last_save_time
+        if self._interval < elapsed + self._postpone_delay:
+            self._interval = elapsed + self._postpone_delay
+            if self._interval > self._interval_max:
+                self._interval = self._interval_max
+        _logger.debug("postponing autosave: current interval {}, "
+                      "elapsed since last save {}"
+                      .format(self._interval, elapsed))
 
     def _on_timer(self):
-        t = time.time()
-        if t - self._last_save_time > self._interval:
-            self._last_save_time = t
+        now = time.time()
+        elapsed = now - self._last_save_time
+        if self._interval < elapsed and \
+           self._pause == 0:
+            self._last_save_time = now
+            self._interval = self._interval_min
+            _logger.debug("auto-saving language models; "
+                          "interval {}, elapsed time {}"
+                          .format(self._interval, elapsed))
             self._model_cache.save_models()
-        return True # run again
+
+        if self._pause:
+            self._pause = max(0, self._pause - self._timer_interval)
+
+        return True  # run again
+
 
