@@ -1,7 +1,10 @@
+
 /*
- * Copyright © 2006, 2007, 2008 Chris Jones
- * Copyright © 2010 marmuta
- * Copyright © 2013 Gerd Kohlberger
+ * Copyright © 2006-2008 Chris Jones <tortoise@tortuga>
+ * Copyright © 2010, 2013 marmuta <marmvta@gmail.com>
+ * Copyright © 2013 Gerd Kohlberger <lowfi@chello.at>
+ *
+ * This file is part of Onboard.
  *
  * Onboard is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -1362,10 +1365,6 @@ vk_get_label_from_keysym (KeySym keyval)
             label = PyString_FromString(".");
             break;
 
-        case GDK_KEY_horizconnector:
-            label = PyString_FromString("horiz\nconn");
-            break;
-
         case GDK_KEY_Mode_switch:
             label = PyString_FromString("AltGr");
             break;
@@ -1386,16 +1385,37 @@ vk_get_label_from_keysym (KeySym keyval)
             else
             {
                 const char *name = gdk_keyval_name (keyval);
-                if (name)
-                    label = PyString_FromStringAndSize(name, 2);
-                else
+                if (!name)		
+                {
                     label = PyString_FromString("");
+                }
+                else
+                {
+                    size_t l = strlen(name);
+                    if (l > 2 && name[0] == '0' && name[1] == 'x') // hex number?
+                    {
+                        // Most likely an erroneous keysym and not Onboard's fault.
+                        // Show the hex number fully as label for easier debugging.
+                        // Happend with Belgian CapsLock+keycode 51,
+                        // keysym 0x039c on Xenial.
+                        label = PyString_FromStringAndSize(name, MIN(l, 10));
+                    }
+                    else
+                    {
+                        label = PyString_FromStringAndSize(name, MIN(l, 2));
+                    }
+                }
             }
         }
     }
     return label;
 }
 
+// Don't use const int for the max label size, else stack smashing error
+// with -fstack-protector and GCC 4.6.3 on Precise. Later versions
+// are unaffected.
+#define max_label_size 128
+
 static PyObject *
 osk_virtkey_labels_from_keycode (PyObject *self, PyObject *args)
 {
@@ -1406,8 +1426,13 @@ osk_virtkey_labels_from_keycode (PyObject *self, PyObject *args)
     Py_ssize_t n;
     KeyCode keycode;
     KeySym keysym = 0;
-    unsigned int mods_rtn;
     int i, group;
+    unsigned int mods;
+    unsigned int mods_rtn;
+
+    XKeyPressedEvent ev;
+    char label[max_label_size+1];
+    int label_size = 0;
 
     if (!PyArg_ParseTuple (args, "i|O", &keycode, &omod_masks))
         return NULL;
@@ -1426,8 +1451,13 @@ osk_virtkey_labels_from_keycode (PyObject *self, PyObject *args)
         n = PySequence_Fast_GET_SIZE (seq);
         ret = PyTuple_New (n);
 
+        memset(&ev, 0, sizeof(ev));
+        ev.type = KeyPress;
+        ev.display = vk->display;
+
         for (i = 0; i < n; i++, items++)
         {
+
             if (!PyLong_Check (*items))
             {
                 PyErr_SetString (PyExc_ValueError, "expected integer");
@@ -1436,15 +1466,18 @@ osk_virtkey_labels_from_keycode (PyObject *self, PyObject *args)
                 return NULL;
             }
 
-            if (XkbTranslateKeyCode (vk->kbd,
-                                     keycode,
-                                     XkbBuildCoreState (PyLong_AsLong (*items), group),
-                                     &mods_rtn,
-                                     &keysym))
+            mods = XkbBuildCoreState (PyLong_AsLong (*items), group);
+
+            ev.state = mods;
+            ev.keycode = keycode;
+            label_size = XLookupString(&ev, label, max_label_size, &keysym, NULL);
+            label[label_size] = '\0';
+
+            if (keysym)
                 PyTuple_SET_ITEM (ret, i, vk_get_label_from_keysym (keysym));
             else
-                PyTuple_SET_ITEM (ret, i, PyString_FromString (""));
-        }
+                PyTuple_SET_ITEM (ret, i, PyString_FromString (label));
+       }
         Py_DECREF (seq);
     }
     else
@@ -1467,6 +1500,8 @@ osk_virtkey_labels_from_keycode (PyObject *self, PyObject *args)
     return ret;
 }
 
+#undef max_label_size
+
 static PyObject *
 osk_virtkey_keysyms_from_keycode (PyObject *self, PyObject *args)
 {
