@@ -1,7 +1,27 @@
 #!/usr/bin/python3
+
+# Copyright © 2009 Chris Jones <tortoise@tortuga>
+# Copyright © 2011 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2011-2014 marmuta <marmvta@gmail.com>
 #
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
+
 # This file contains a copy of the strings of the layout files (.onboard files)
-# that needs translation. As the current build system cannot extract the
+# that need translation. As the current build system cannot extract the
 # strings that need translation directly from the layout files, it gets them
 # from this file.
 #
@@ -277,7 +297,10 @@ TRANSLATABLE_LAYOUT_STRINGS = [
     # translators: this is a 'tooltip' of the key_template 'settings' in keyboard layout 'key_defs.xml'
     _("Preferences"),
 
-    # translators: this is a 'tooltip' of the key 'quit' in keyboard layout 'Whiteboard_wide.onboard'
+    # translators: this is a 'tooltip' of the key 'previous-predictions.wordlist' in keyboard layout 'word_suggestions.xml'
+    _("Previous suggestions"),
+
+    # translators: this is a 'tooltip' of the key 'quit' in keyboard layout 'Whiteboard.onboard'
     _("Quit Keyboard"),
 
     # translators: this is a 'tooltip' of the key_template 'quit' in keyboard layout 'key_defs.xml'
@@ -323,7 +346,7 @@ TRANSLATABLE_LAYOUT_STRINGS = [
     _("Special Characters"),
 
     # translators: this is a 'summary' of the keyboard layout 'Whiteboard_wide.onboard'
-    _("Special characters for interactive Whiteboards, wide"),
+    _("Special characters for interactive whiteboards, wide"),
 
     # translators: this is a 'summary' of the keyboard layout 'Whiteboard.onboard'
     _("Special characters for interactive whiteboards"),
